"""This module contains the base class of all music providers."""

from __future__ import annotations

from typing import Optional

from core.musiq import musiq, playback
from core.settings import storage
from core.tasks import app


class ProviderError(Exception):
    """An error to indicate that an error occurred while providing music."""


class WrongUrlError(Exception):
    """An error to indicate that a provider was called
    with a url that belongs to a different service."""


class MusicProvider:
    """The base class for all music providers.
    Provides abstract function declarations."""

    def __init__(self, query: Optional[str], key: Optional[int]) -> None:
        self.query = query
        self.key = key
        if not hasattr(self, "type"):
            # the type should already have been set by the base class
            self.type = "unknown"
            assert False
        self.id: Optional[str] = None
        self.ok_message = "ok"
        self.error = "error"

    def check_cached(self) -> bool:
        """Returns whether this resource is available on disk."""
        raise NotImplementedError()

    def check_available(self) -> bool:
        """Returns whether this resource is available online."""
        raise NotImplementedError()

    def enqueue_placeholder(self, manually_requested) -> None:
        """Enqueues a placeholder if applicable. Playlists have no placeholder, only songs do.
        Used to identify this resource in the client after a request."""
        raise NotImplementedError()

    def remove_placeholder(self) -> None:
        """Removes the placeholder in the queue that represents this resource.
        Called if there was an error and this element needs to be removed from the queue.
        """
        raise NotImplementedError()

    def make_available(self) -> bool:
        """Makes this resource available for playback.
        If possible, downloads it to disk.
        If this takes a long time, calls update_state so the placeholder is visible.
        Returns False if an error occured, True otherwise."""
        raise NotImplementedError()

    def was_requested_before(self) -> bool:
        """Returns whether this resource has been manually requested at least once."""
        raise NotImplementedError()

    def on_cooldown(self) -> bool:
        """Returns whether this resource is currently on cooldown and may not be requested.
        Sets the provider error."""
        raise NotImplementedError()

    def persist(self, session_key: str, archive: bool = True) -> None:
        """Updates the database.
        Creates an archived entry or updates it.
        Also handles logging to database."""
        raise NotImplementedError()

    def enqueue(self) -> None:
        """Updates the placeholder in the song queue with the actual data."""
        raise NotImplementedError()

    def request(
        self, session_key: str, archive: bool = True, manually_requested: bool = True
    ) -> None:
        """Tries to request this resource.
        Uses the local cache if possible, otherwise tries to retrieve it online."""

        if 0 < storage.get("max_queue_length") <= playback.queue.count():
            self.error = "Queue limit reached"
            raise ProviderError(self.error)

        enqueue_function = enqueue

        if not self.check_cached():
            if self.query is not None and storage.get("additional_keywords"):
                # add the additional keywords from the settings before checking
                self.query += " " + storage.get("additional_keywords")
            if not self.check_available():
                raise ProviderError(self.error)

            # overwrite the enqueue function and make the resource available before calling it
            enqueue_function = fetch_enqueue

        if storage.get("new_music_only"):
            if self.was_requested_before():
                self.error = "Only new music is allowed!"
                raise ProviderError(self.error)

        if storage.get("song_cooldown") != 0:
            if self.on_cooldown():
                raise ProviderError(self.error)

        self.enqueue_placeholder(manually_requested)

        enqueue_function.delay(self, session_key, archive)


@app.task
def enqueue(provider: MusicProvider, session_key: str, archive: bool) -> None:
    """Enqueue the music managed by the given provider."""
    provider.persist(session_key, archive=archive)
    provider.enqueue()


@app.task
def fetch_enqueue(provider: MusicProvider, session_key: str, archive: bool) -> None:
    """Fetch and enqueue the music managed by the given provider."""
    if not provider.make_available():
        provider.remove_placeholder()
        musiq.update_state()
        return

    enqueue(provider, session_key, archive)



"""This module handles the screen configuration."""
import math
import re
import subprocess
import os
from typing import List, Tuple, cast

from django.conf import settings as conf
from core import redis, util
from core.lights import lights
from core.lights.device import Device
from core.settings import storage


class Screen(Device):
    """This class provides an interface to control the screen."""

    @staticmethod
    def get_primary() -> str:
        """Return the primary screen"""
        for line in subprocess.check_output(
            "xrandr -q".split(), text=True
        ).splitlines():
            output = re.match(r"(\S+) connected primary", line)
            if output:
                return output.group(1)
        raise ValueError("Could not find primary screen")

    def __init__(self, manager) -> None:
        super().__init__(manager, "screen")

        # set the DISPLAY environment variable the correct X Display is used
        os.environ["DISPLAY"] = ":0"
        # the visualization needs X to work, so we check if it is running
        try:
            subprocess.check_call(
                "xset q".split(), stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
            )
        except (FileNotFoundError, subprocess.CalledProcessError):
            # Cannot connect to X
            return

        # disable blanking and power saving
        subprocess.call("xset s off".split())
        subprocess.call("xset s noblank".split())
        subprocess.call("xset -dpms".split())

        # ignore the scale factor for large displays,
        # we always render fullscreen without scaling
        os.environ["WINIT_X11_SCALE_FACTOR"] = "1"

        # method should additionally check whether hdmi is connected
        # however, I found no way to do that
        # without hdmi_force_hotplug=1:
        # tvservice -M gives attached events, but tvserice -s is always connected
        # hdmi cannot be plugged in after boot
        # with hdmi_force_hotplug=1:
        # tvservice -M records nothing, tvserice -s is always connected
        # /sys/class/drm/card1-HDMI-A-1/status is always connected
        #
        # so we set hotplug and initialize the screen even if none is connected
        self.initialized = True
        redis.put("screen_initialized", True)

        self.output = self.get_primary()
        self.resolution = (0, 0)  # set in adjust
        self.adjust()

    def adjust(self) -> None:
        """Updates resolutions and resets the current one.
        Needed after changing screens or hotplugging after booting without a connected screen."""
        self.resolution = storage.get("initial_resolution")
        resolutions = list(reversed(sorted(self.list_resolutions())))
        redis.put("resolutions", resolutions)
        # if unset, initialize with the highest resolution
        if self.resolution == (0, 0):
            storage.put("initial_resolution", resolutions[0])
            self.resolution = resolutions[0]
        self.set_resolution(self.resolution)

    def clear(self) -> None:
        # when no program is running, nothing is shown on screen
        pass

    def list_resolutions(self) -> List[Tuple[int, int]]:
        """Returns all supported resolutions that match the preferred resolution in aspect ratio."""
        if not self.initialized:
            return []
        modes = []
        listing_modes = False
        preferred_mode = (1280, 720)
        for line in subprocess.check_output(
            "xrandr -q".split(), text=True
        ).splitlines():
            output = re.match(r"(\S+) connected", line)
            if output:
                listing_modes = output.group(1) == self.output
            if not listing_modes:
                continue
            match = re.match(r"\s+(\d+x\d+)\s+\d+\.\d+", line)
            if match:
                mode = tuple(map(int, match.group(1).split("x")))
                mode = cast(Tuple[int, int], mode)
                modes.append(mode)
                if "+" in line:
                    preferred_mode = mode
        preferred_ratio = preferred_mode[0] / preferred_mode[1]
        usable_modes = []
        for mode in modes:
            ratio = mode[0] / mode[1]
            if math.isclose(ratio, preferred_ratio, rel_tol=0.1):
                usable_modes.append(mode)
        return usable_modes

    def set_resolution(self, resolution: Tuple[int, int]) -> None:
        """Sets the current output to the given resolution.
        Also updates the background in production."""
        if not self.initialized:
            return
        subprocess.call(
            [
                "xrandr",
                "--output",
                self.output,
                "--mode",
                util.format_resolution(resolution),
            ]
        )

        if not conf.DEBUG:
            # show a background that is visible when the visualization is not rendering
            # needs to be reset every resolution change
            try:
                subprocess.call(
                    [
                        "feh",
                        "--bg-max",
                        os.path.join(conf.BASE_DIR, "resources/images/background.png"),
                    ]
                )
            except FileNotFoundError:
                pass

        redis.put("current_resolution", resolution)
        self.resolution = resolution
        lights.update_state()

    def lower_resolution(self) -> None:
        """Sets the resolution to the next resolution with similar ratio that is smaller."""
        if not self.initialized:
            return
        resolutions = redis.get("resolutions")
        index = resolutions.index(self.resolution)
        index = min(index + 1, len(resolutions) - 1)
        self.set_resolution(resolutions[index])



"""This file contains all exceptions related to this module."""


class ScreenProgramStopped(Exception):
    """Raised whenever a running screen program stopped execution.
    Usually through external reasons like closing the window."""



from core.musiq import player


class FakePlayer(player.Player):
    def start_song(self, song, catch_up: float):
        pass

    def play_alarm(self, interrupt: bool, alarm_path: str) -> None:
        pass

    def play_backup_stream(self):
        pass


def restart() -> None:
    pass


def seek_backward(_seek_distance: float) -> None:
    pass


def play() -> None:
    pass


def pause() -> None:
    pass


def seek_forward(_seek_distance: float) -> None:
    pass


def skip() -> None:
    pass


def set_volume(_volume) -> None:
    pass



"""This module manages and counts user accesses and handles permissions."""
import re
from ast import literal_eval
import colorsys
import random
import time
from functools import wraps
from typing import Callable, Optional

import ipware
from django.contrib.sessions.models import Session
from django.core.handlers.wsgi import WSGIRequest
from django.db import transaction
from django.http import HttpResponse, HttpResponseBadRequest
from django.utils import timezone

from core import redis

# kick users after some time without any request
from core.lights import leds
from core.settings import storage
from core.settings.storage import Privileges
from core.util import extract_value

INACTIVITY_PERIOD = 600


def has_controls(user) -> bool:
    """Determines whether the given user is allowed to control playback."""
    return user.username == "mod" or is_admin(user)


def is_admin(user) -> bool:
    """Determines whether the given user is the admin."""
    return user.is_superuser


def has_privilege(user, privilege: Privileges):
    if privilege == Privileges.everybody:
        return True
    elif privilege == Privileges.mod and has_controls(user):
        return True
    elif privilege == Privileges.admin and is_admin(user):
        return True
    return False


def update_user_count() -> None:
    """Go through all recent requests and delete those that were too long ago."""
    now = time.time()
    last_requests = redis.get("last_requests")
    for key, value in list(last_requests.items()):
        if now - value >= INACTIVITY_PERIOD:
            del last_requests[key]
            redis.put("last_requests", last_requests)
    redis.put("last_user_count_update", now)


def get_count() -> int:
    """Returns the number of currently active users.
    Updates this number after an intervals since the last update."""
    if time.time() - redis.get("last_user_count_update") >= 60:
        update_user_count()
    return len(redis.get("last_requests"))


def partymode_enabled() -> bool:
    """Determines whether partymode is enabled,
    based on the number of currently active users."""
    return len(redis.get("last_requests")) >= storage.get("people_to_party")


def get_client_ip(request: WSGIRequest):
    """Returns the origin IP of a given request or "" if not possible."""
    request_ip, _ = ipware.get_client_ip(request)
    if request_ip is None:
        request_ip = ""
    return request_ip


def try_vote(request_ip: str, queue_key: int, amount: int) -> bool:
    """If the user can not vote any more for the song into the given direction, return False.
    Otherwise, perform the vote and returns True."""
    # Votes are stored as individual (who, what) tuples in redis.
    # A mapping who -> [what, ...] is not used,
    # because each modification would require deserialization and subsequent serialization.
    # Without such a mapping we cannot easily find all votes belonging to a session key,
    # which would be required to update the view of a user whose client-votes got desynced.
    # This should never happen during normal usage, so we optimize for our main use case:
    # looking up whether a single user voted for a single song, which is constant with tuples.
    # Since this feature indexes by the request IP and not the session_key,
    # it can not share its data structure with the votes for the color indicators.
    entry = str((request_ip, queue_key))
    allowed = True

    # redis transaction: https://github.com/Redis/redis-py#pipelines
    def check_entry(pipe) -> None:
        nonlocal allowed
        vote = pipe.get(entry)
        if vote is None:
            new_vote = amount
        else:
            new_vote = int(vote) + amount
        if new_vote < -1 or new_vote > 1:
            allowed = False
            return
        allowed = True
        # expire these entries to avoid accumulation over long runtimes.
        pipe.multi()
        pipe.set(entry, new_vote, ex=24 * 60 * 60)

    redis.connection.transaction(check_entry, entry)
    return allowed


def _get_next_color():
    with transaction.atomic():
        next_index = storage.get("next_color_index")
        storage.put("next_color_index", next_index + 1)

    offset = storage.get("color_offset")

    hue = offset + next_index * (137.508 / 360)  # approximation for the golden angle
    color = colorsys.hsv_to_rgb(hue, 0.4, 1)

    return "#%02x%02x%02x" % tuple(round(v * 255) for v in color)


def color_of(session_key: str) -> Optional[str]:
    if not session_key or storage.get("color_indication") == storage.Privileges.nobody:
        return None
    # no transaction because this is called many times and at worst race conditions would result
    # in a different color being set
    color = redis.connection.get("color-" + session_key)
    if color is None:
        color = _get_next_color()
    # TODO: this is lost on server restart.
    # maybe store the color client side so it can be recovered?
    redis.connection.set("color-" + session_key, color, ex=24 * 60 * 60)
    return color


def register_song(request: WSGIRequest, queue_key: int) -> None:
    # For each song, identified by its queue_key, the following information is stored:
    # (<session_key>, {<session_key>: <vote>, …})
    # This requires a session_key, thus it can only be used in @tracked functions
    session_key = request.session.session_key

    key = f"engagement-{queue_key}"

    def update_entry(pipe) -> None:
        value = pipe.get(key)
        engagement = None if value is None else literal_eval(value)
        if engagement is None:
            engagement = (None, {})
        _, votes = engagement
        # expire these entries to avoid accumulation over long runtimes.
        pipe.multi()
        pipe.set(key, str((session_key, votes)), ex=24 * 60 * 60)

    redis.connection.transaction(update_entry, key)


def register_vote(request: WSGIRequest, queue_key: int, amount: int) -> None:
    session_key = request.session.session_key

    key = f"engagement-{queue_key}"

    def update_entry(pipe) -> None:
        value = pipe.get(key)
        engagement = None if value is None else literal_eval(value)
        if engagement is None:
            engagement = (None, {})
        requested_by, votes = engagement
        if session_key in votes:
            votes[session_key] += amount
        else:
            votes[session_key] = amount
        # clamp votes to [-1,1]. This helps recovering from desyncs after redis was cleared
        # but client votes are still locked in
        votes[session_key] = max(-1, min(1, votes[session_key]))
        if votes[session_key] == 0:
            del votes[session_key]
        # expire these entries to avoid accumulation over long runtimes.
        pipe.multi()
        pipe.set(key, str((requested_by, votes)), ex=24 * 60 * 60)

    redis.connection.transaction(update_entry, key)


def set_user_color(request: WSGIRequest) -> HttpResponse:
    """Updates the color assigned to the session of the request."""
    from core.musiq import musiq

    color, response = extract_value(request.POST)
    if not color or not re.match(r"^#[0-9a-f]{6}$", color):
        return HttpResponseBadRequest()
    session_key = request.session.session_key
    redis.connection.set("color-" + session_key, color, ex=24 * 60 * 60)
    musiq.update_state()
    return response


def tracked(
    func: Callable[[WSGIRequest], HttpResponse]
) -> Callable[[WSGIRequest], HttpResponse]:
    """A decorator that stores the last access for every connected ip
    so the number of active users can be determined."""

    def _decorator(request: WSGIRequest) -> HttpResponse:
        # create a sessions if none exists (necessary for anonymous users)
        if not request.session or not request.session.session_key:
            # if there are no active sessions (= this is the first one)
            # reset the color index and choose a new offset.
            active_sessions = Session.objects.filter(
                expire_date__gte=timezone.now()
            ).count()
            if active_sessions == 0:
                storage.put("color_offset", random.random())
                storage.put("next_color_index", 0)

            request.session.save()

        request_ip = get_client_ip(request)
        last_requests = redis.get("last_requests")
        last_requests[request_ip] = time.time()
        redis.put("last_requests", last_requests)

        def check():
            active = redis.get("active_requests")
            if active > 0:
                leds.enable_act_led()
            else:
                leds.disable_act_led()

        redis.connection.incr("active_requests")
        check()
        response = func(request)
        redis.connection.decr("active_requests")
        check()

        return response

    return wraps(func)(_decorator)



"""This module provides common functionality for all pages on the site."""

import os
import random
from typing import Any, Dict

from django.core.handlers.wsgi import WSGIRequest
from django.db import transaction
from django.http import HttpResponseBadRequest, HttpResponseRedirect
from django.http.response import HttpResponse, JsonResponse
from django.shortcuts import render
from django.urls import reverse

from django.conf import settings as conf
from core import models, redis, user_manager
from core.lights import controller
from core.musiq import musiq
from core.settings import storage
from core.settings import system
from core.state_handler import send_state


def _get_random_hashtag() -> str:
    active_hashtags = models.Tag.objects.filter(active=True)
    if active_hashtags.count() == 0:
        return "Add your first hashtag!"
    index = random.randint(0, active_hashtags.count() - 1)
    hashtag = active_hashtags[index]
    return hashtag.text


def _get_apk_link() -> str:
    local_apk = os.path.join(conf.STATIC_FILES, "apk/shareberry.apk")
    if os.path.isfile(local_apk):
        assert conf.STATIC_URL
        return os.path.join(conf.STATIC_URL, "apk/shareberry.apk")
    return "https://github.com/raveberry/shareberry/releases/latest/download/shareberry.apk"


def _increment_counter() -> int:
    with transaction.atomic():
        counter = models.Counter.objects.get_or_create(id=1, defaults={"value": 0})[0]
        counter.value += 1
        counter.save()
    update_state()
    return counter.value


def context(request: WSGIRequest) -> Dict[str, Any]:
    """Returns the base context that is needed on every page.
    Increments the visitors counter."""
    from core import urls

    _increment_counter()
    return {
        "base_urls": urls.base_paths,
        "interactivity": storage.get("interactivity"),
        "interactivities": {
            "fullControl": storage.Interactivity.full_control,
            "fullVoting": storage.Interactivity.full_voting,
            "upvotesOnly": storage.Interactivity.upvotes_only,
            "noControl": storage.Interactivity.no_control,
        },
        "color_indication": user_manager.has_privilege(
            request.user, storage.get("color_indication")
        ),
        # We pass the color to the user here for two reasons
        # 1. a color is assigned to every user on page load, not on first interaction
        # 2. during state updates every session receives all information,
        #    thus we would have to send the mapping to everyone, revealing session keys
        "user_color": user_manager.color_of(request.session.session_key),
        "privileges": {
            "everybody": storage.Privileges.everybody,
            "mod": storage.Privileges.mod,
            "admin": storage.Privileges.admin,
            "nobody": storage.Privileges.nobody,
        },
        "hashtag": _get_random_hashtag(),
        "demo": conf.DEMO,
        "controls_enabled": user_manager.has_controls(request.user)
        or storage.get("interactivity") == storage.Interactivity.full_control,
        "is_admin": user_manager.is_admin(request.user),
        "apk_link": _get_apk_link(),
        "local_enabled": storage.get("local_enabled"),
        "youtube_enabled": storage.get("youtube_enabled"),
        "spotify_enabled": storage.get("spotify_enabled"),
        "soundcloud_enabled": storage.get("soundcloud_enabled"),
        "jamendo_enabled": storage.get("jamendo_enabled"),
    }


def state_dict() -> Dict[str, Any]:
    """This function constructs a base state dictionary with website wide state.
    Pages sending states extend this state dictionary."""
    try:
        default_platform = musiq.enabled_platforms_by_priority()[0]
    except IndexError:
        default_platform = ""
    return {
        "partymode": user_manager.partymode_enabled(),
        "users": user_manager.get_count(),
        "visitors": models.Counter.objects.get_or_create(id=1, defaults={"value": 0})[
            0
        ].value,
        "lightsEnabled": redis.get("lights_active"),
        "playbackError": redis.get("playback_error"),
        "alarm": redis.get("alarm_playing"),
        "defaultPlatform": default_platform,
    }


def no_stream(request: WSGIRequest) -> HttpResponse:
    """Renders the /stream page. If this is reached, there is no stream active."""
    return render(request, "no_stream.html", context(request))


def submit_hashtag(request: WSGIRequest) -> HttpResponse:
    """Add the given hashtag to the database."""
    hashtag = request.POST.get("hashtag")
    if hashtag is None or len(hashtag) == 0:
        return HttpResponseBadRequest()

    if hashtag[0] != "#":
        hashtag = "#" + hashtag
    models.Tag.objects.create(text=hashtag, active=storage.get("hashtags_active"))

    return HttpResponse()


def logged_in(request: WSGIRequest) -> HttpResponse:
    """This endpoint is visited after every login.
    Redirect the admin to the settings and everybody else to the musiq page."""
    if user_manager.is_admin(request.user):
        return HttpResponseRedirect(reverse("settings"))
    return HttpResponseRedirect(reverse("musiq"))


def set_user_color(request: WSGIRequest) -> None:
    """Set user color for indication of votes.
    Situated in base because the dropdown is accessible from every page."""
    return user_manager.set_user_color(request)


def set_lights_shortcut(request: WSGIRequest) -> None:
    """Request endpoint for the lights shortcut.
    Situated in base because the dropdown is accessible from every page."""
    return controller.set_lights_shortcut(request)


def upgrade_available(_request: WSGIRequest) -> HttpResponse:
    """Checks whether newer Raveberry version is available."""
    latest_version = system.fetch_latest_version()
    current_version = conf.VERSION
    if latest_version and latest_version != current_version:
        return JsonResponse(True, safe=False)
    return JsonResponse(False, safe=False)


def update_state() -> None:
    """Sends an update event to all connected clients."""
    send_state(state_dict())



"""This module defines the routing for realtime websocket requests."""

from django.urls import path

from core import state_handler

WEBSOCKET_URLPATTERNS = [path("state/", state_handler.StateConsumer.as_asgi())]



"""This module contains all Youtube related code."""
# We need to access yt-dlp's internal methods for some features
# pylint: disable=protected-access

from __future__ import annotations

import errno
import logging
import os
import pickle
import subprocess
import urllib.parse
from contextlib import contextmanager
from typing import Any, Dict, Iterator, List, Optional, cast
from urllib.parse import parse_qs, urlparse

import requests
import yt_dlp
import ytmusicapi
from django.conf import settings
from django.http.response import HttpResponse

from core.musiq import musiq, song_utils
from core.musiq.playlist_provider import PlaylistProvider
from core.musiq.song_provider import SongProvider
from core.settings import storage


@contextmanager
def youtube_session() -> Iterator[requests.Session]:
    """This context opens a requests session and loads the youtube cookies file."""

    cookies_path = os.path.join(settings.BASE_DIR, "config/youtube_cookies.pickle")
    session = requests.session()
    # Have yt-dlp deal with consent cookies etc to setup a valid session
    extractor = yt_dlp.extractor.youtube.YoutubeIE()
    extractor._downloader = yt_dlp.YoutubeDL()
    extractor.initialize()
    session.cookies.update(extractor._downloader.cookiejar)

    try:
        if os.path.getsize(cookies_path) > 0:
            with open(cookies_path, "rb") as cookies_file:
                session.cookies.update(pickle.load(cookies_file))
    except FileNotFoundError:
        pass

    headers = {"User-Agent": yt_dlp.utils.random_user_agent()}
    session.headers.update(headers)
    yield session

    with open(cookies_path, "wb") as cookies_file:
        pickle.dump(session.cookies, cookies_file)


class YoutubeDLLogger:
    """This logger class is used to log process of yt-dlp downloads."""

    @classmethod
    def debug(cls, msg: str) -> None:
        """This method is called if yt-dlp does debug level logging."""
        logging.debug(msg)

    @classmethod
    def warning(cls, msg: str) -> None:
        """This method is called if yt-dlp does warning level logging."""
        logging.debug(msg)

    @classmethod
    def error(cls, msg: str) -> None:
        """This method is called if yt-dlp does error level logging."""
        logging.error(msg)


class Youtube:
    """This class contains code for both the song and playlist provider"""

    used_info_dict_keys = {"id", "filesize", "url", "_type", "title", "entries"}

    @staticmethod
    def get_ydl_opts() -> Dict[str, Any]:
        """This method returns a dictionary containing sensible defaults for yt-dlp options.
        It is roughly equivalent to the following command:
        yt-dlp --format bestaudio[ext=m4a]/best[ext=m4a] --output '%(id)s.%(ext)s' \
            --no-playlist --no-cache-dir --write-thumbnail --default-search auto \
            --add-metadata --embed-thumbnail
        """
        postprocessors = [
            {"key": "FFmpegMetadata"},
            {
                "key": "EmbedThumbnail",
                # overwrite any thumbnails already present
                "already_have_thumbnail": True,
            },
        ]
        return {
            "format": "bestaudio[ext=m4a]/best[ext=m4a]",
            "outtmpl": os.path.join(settings.SONGS_CACHE_DIR, "%(id)s.%(ext)s"),
            "noplaylist": True,
            "cachedir": False,
            "no_color": True,
            "writethumbnail": True,
            "default_search": "auto",
            "postprocessors": postprocessors,
            "logger": YoutubeDLLogger(),
        }

    @staticmethod
    def get_search_suggestions(query: str) -> List[str]:
        """Returns a list of suggestions for the given query from Youtube."""
        with youtube_session() as session:
            params = {
                "client": "youtube",
                "q": query[:100],  # queries longer than 100 characters are not accepted
                "xhr": "t",  # this makes the response be a json file
            }
            response = session.get(
                "https://clients1.google.com/complete/search", params=params
            )
        suggestions = ytmusicapi.YTMusic().get_search_suggestions(query)
        try:
            if suggestions[0] == query:
                suggestions = suggestions[1:]
        except IndexError:
            return []
        return suggestions


class YoutubeSongProvider(SongProvider, Youtube):
    """This class handles songs from Youtube."""

    @staticmethod
    def get_id_from_external_url(url: str) -> str:
        return parse_qs(urlparse(url).query)["v"][0]

    def __init__(self, query: Optional[str], key: Optional[int]) -> None:
        self.type = "youtube"
        super().__init__(query, key)
        self.info_dict: Dict[str, Any] = {}

    def check_cached(self) -> bool:
        if not self.id:
            # id could not be extracted from query, needs to be serched
            return False
        if storage.get("output") == "client":
            # youtube streaming links need to be fetched each time the song is requested
            return False
        return os.path.isfile(self.get_path())

    def check_available(self) -> bool:
        info_dict = None

        def extract_info(id):
            nonlocal info_dict
            try:
                with yt_dlp.YoutubeDL(Youtube.get_ydl_opts()) as ydl:
                    info_dict = ydl.extract_info(id, download=False)
                    return True
            except (yt_dlp.utils.ExtractorError, yt_dlp.utils.DownloadError) as error:
                logging.warning("error during availability check for %s:", id)
                logging.warning(error)
            return False

        if self.id:
            # do not search if an id is already present
            extract_info(self.id)
        else:
            # do not filter to only receive "song" results, because we would skip the top result
            results = ytmusicapi.YTMusic().search(self.query)
            for result in results:
                if result["resultType"] not in ("video", "song"):
                    continue
                if song_utils.is_forbidden(result["title"]):
                    continue
                if extract_info(result["videoId"]):
                    break

        if not info_dict:
            self.error = "No songs found"
            return False

        self.id = info_dict["id"]

        return self.check_not_too_large(info_dict["filesize"])

    def _download(self) -> bool:
        download_error = None
        location = None

        try:
            with yt_dlp.YoutubeDL(Youtube.get_ydl_opts()) as ydl:
                ydl.download([self.get_external_url()])

            location = self.get_path()
            base = os.path.splitext(location)[0]
            thumbnail = base + ".jpg"
            try:
                os.remove(thumbnail)
            except FileNotFoundError:
                logging.info("tried to delete %s but does not exist", thumbnail)

            try:
                # tag the file with replaygain to perform volume normalization
                subprocess.call(
                    ["rganalysis", location],
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
            except OSError as error:
                if error.errno == errno.ENOENT:
                    pass  # the rganalysis package was not found. Skip normalization
                else:
                    raise

        except yt_dlp.utils.DownloadError as error:
            download_error = error

        if download_error is not None or location is None:
            logging.error("accessible video could not be downloaded: %s", self.id)
            logging.error("location: %s", location)
            logging.error(download_error)
            return False
        return True

    def make_available(self) -> bool:
        if os.path.isfile(self.get_path()):
            # don't download the file if it is already cached
            return True
        musiq.update_state()
        return self._download()

    def get_path(self) -> str:
        """Return the path in the local filesystem to the cached sound file of this song."""
        if not self.id:
            raise ValueError()
        return song_utils.get_path(self.id + ".m4a")

    def get_internal_url(self) -> str:
        return "file://" + urllib.parse.quote(self.get_path())

    def get_external_url(self) -> str:
        if not self.id:
            raise ValueError()
        return "https://www.youtube.com/watch?v=" + self.id

    def gather_metadata(self) -> bool:
        self.metadata = self.get_local_metadata(self.get_path())
        return True

    def get_suggestion(self) -> str:
        result = ytmusicapi.YTMusic().get_watch_playlist(self.id, limit=2)
        # the first entry usually is the song itself -> use the second one
        suggested_id = result["tracks"][1]["videoId"]
        return "https://www.youtube.com/watch?v=" + suggested_id

    def request_radio(self, session_key: str) -> HttpResponse:
        if not self.id:
            raise ValueError()

        result = ytmusicapi.YTMusic().get_watch_playlist(
            self.id, limit=storage.get("max_playlist_items"), radio=True
        )
        radio_id = result["playlistId"]

        provider = YoutubePlaylistProvider("", None)
        provider.id = radio_id
        provider.title = radio_id
        for entry in result["tracks"]:
            provider.urls.append("https://www.youtube.com/watch?v=" + entry["videoId"])
        provider.request("", archive=False, manually_requested=False)
        return HttpResponse("queueing radio (might take some time)")


class YoutubePlaylistProvider(PlaylistProvider, Youtube):
    """This class handles Youtube Playlists."""

    @staticmethod
    def get_id_from_external_url(url: str) -> Optional[str]:
        try:
            list_id = parse_qs(urlparse(url).query)["list"][0]
        except KeyError:
            return None
        return list_id

    def __init__(self, query: Optional[str], key: Optional[int]) -> None:
        self.type = "youtube"
        super().__init__(query, key)

    def is_radio(self) -> bool:
        if not self.id:
            raise ValueError()
        return self.id.startswith("RD")

    def search_id(self) -> Optional[str]:
        results = ytmusicapi.YTMusic().search(self.query)

        for result in results:
            if result["resultType"] not in (
                "playlist",
                "community_playlist",
                "featured_playlist",
            ):
                continue
            if "browseId" not in result or not result["browseId"]:
                continue
            # remove the preceding "VL" from the playlist id
            list_id = result["browseId"][2:]
            return list_id

    def fetch_metadata(self) -> bool:
        assert self.id

        # radio playlists are prefilled when requesting them
        if self.title and self.urls:
            return True

        try:
            result = ytmusicapi.YTMusic().get_playlist(self.id)
        except Exception as e:
            # query was not a playlist url -> search for the query
            assert False

        assert self.id == result["id"]
        self.title = result["title"]
        for entry in result["tracks"]:
            if "videoId" not in entry or not entry["videoId"]:
                continue
            self.urls.append("https://www.youtube.com/watch?v=" + entry["videoId"])
        assert self.key is None

        return True



"""This module contains the syncsongmetadata command."""
from django.core.management.base import BaseCommand

from core.musiq import song_utils


class Command(BaseCommand):
    """Defines the syncsongmetadata command."""

    help = (
        "Goes through every archived song and syncs its metadata from the file system."
    )

    def handle(self, *args, **options):
        from core.models import ArchivedSong
        from core.musiq.song_provider import SongProvider

        for song in ArchivedSong.objects.all():
            try:
                provider = SongProvider.create(external_url=song.url)
            except NotImplementedError:
                # For this song a provider is necessary that is not available
                # e.g. the song was played before, but the provider was disabled
                continue
            cached = provider.check_cached()
            if cached:
                # sync the metadata in the database with the file system
                # _get_path is defined for localdrive and youtube,
                # the only two providers that may be cached
                from core.musiq.local import LocalSongProvider
                from core.musiq.youtube import YoutubeSongProvider

                assert isinstance(provider, (YoutubeSongProvider, LocalSongProvider))
                metadata = song_utils.get_metadata(provider.get_path())
                song.artist = metadata["artist"]
                song.title = metadata["title"]
                song.duration = metadata["duration"]
                song.cached = True
            else:
                # keep old data but store that the song is not cached
                song.cached = False
            song.save()



# Generated by Django 4.0 on 2021-12-09 09:51

import django.contrib.postgres.indexes
from django.db import migrations, connection


class Migration(migrations.Migration):

    dependencies = [("core", "0016_auto_20211128_1900")]

    operations = (
        [
            migrations.AddIndex(
                model_name="archivedquery",
                index=django.contrib.postgres.indexes.GinIndex(
                    django.contrib.postgres.indexes.OpClass("query", "gin_trgm_ops"),
                    name="core_archivedquery_query_trgm",
                ),
            ),
            migrations.AddIndex(
                model_name="archivedsong",
                index=django.contrib.postgres.indexes.GinIndex(
                    django.contrib.postgres.indexes.OpClass("artist", "gin_trgm_ops"),
                    name="core_archivedsong_artist_trgm",
                ),
            ),
            migrations.AddIndex(
                model_name="archivedsong",
                index=django.contrib.postgres.indexes.GinIndex(
                    django.contrib.postgres.indexes.OpClass("title", "gin_trgm_ops"),
                    name="core_archivedsong_title_trgm",
                ),
            ),
            # This value was empirically measured on a Raspberry Pi 4
            # so the suggestion query uses the index, as this is the query that runs most often.
            migrations.RunSQL(
                "ALTER DATABASE raveberry SET random_page_cost=1.4;",
                reverse_sql="ALTER DATABASE raveberry RESET random_page_cost;",
            ),
            # Set word_similarity_threshold to the default value of similarity_threshold.
            # With the default value of 0.6 many typos result in no suggestions.
            migrations.RunSQL(
                "ALTER DATABASE raveberry SET pg_trgm.word_similarity_threshold=0.3;",
                # For some reason there is a permission denied error
                # when trying to reset the threshold, even with owner privileges.
                # Simply do nothing, reversing this parameter is not important.
                reverse_sql="",
            ),
        ]
        if connection.vendor == "postgresql"  # trgm is only available in postgres
        else []
    )



# Generated by Django 2.2.24 on 2021-10-11 18:41

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("core", "0013_auto_20210711_1105")]

    operations = [
        migrations.RenameField(
            model_name="requestlog", old_name="address", new_name="session_key"
        )
    ]



"""This module handles all requests concerning the addition of music to the queue."""

from __future__ import annotations

import ast
import importlib
import logging
import os
from typing import TYPE_CHECKING, Any, Dict, List, Optional, cast

from django.conf import settings as conf
from django.core.handlers.wsgi import WSGIRequest
from django.forms.models import model_to_dict
from django.http import HttpResponseBadRequest
from django.http.response import HttpResponse, JsonResponse
from django.shortcuts import render
from django.utils import timezone
from django.views.decorators.csrf import csrf_exempt

from core import base, redis, user_manager, util
from core.models import CurrentSong, QueuedSong
from core.musiq import controller, playback, song_utils
from core.musiq.music_provider import MusicProvider, ProviderError, WrongUrlError
from core.musiq.playlist_provider import PlaylistProvider
from core.musiq.song_provider import SongProvider
from core.settings import storage
from core.state_handler import send_state
from core.settings.storage import PlatformEnabled, PlatformSuggestions

queue = QueuedSong.objects

if TYPE_CHECKING:
    from core.musiq.song_utils import Metadata


def start() -> None:
    """Initializes the required modules."""

    controller.start()
    playback.start()


def get_alarm_metadata() -> "Metadata":
    """Returns a metadata object for the alarm. The duration is dynamically determined."""
    return {
        "artist": "Raveberry",
        "title": "ALARM!",
        "duration": song_utils.get_metadata(
            os.path.join(conf.BASE_DIR, "resources/sounds/alarm.m4a")
        )["duration"],
        "internal_url": "alarm",
        "external_url": "https://raveberry.party/alarm",
        "stream_url": None,
        "cached": True,
    }


def enabled_platforms_by_priority() -> List[str]:
    """Returns a list of all available platforms, ordered by priority."""
    # local music can only be searched explicitly by key and thus is last
    return [
        platform
        for platform in ["spotify", "youtube", "soundcloud", "jamendo", "local"]
        if storage.get(cast(PlatformEnabled, f"{platform}_enabled"))
    ]


def get_providers(
    query: str,
    key: Optional[int] = None,
    playlist: bool = False,
    preferred_platform: Optional[str] = None,
) -> List[MusicProvider]:
    """Returns a list of all available providers for the given query, ordered by priority.
    If a preferred platform is given, that provider will be first."""

    if key is not None:
        # an archived entry was requested.
        # The key determines the Provider
        provider: MusicProvider
        try:
            if playlist:
                provider = PlaylistProvider.create(query, key)
            else:
                provider = SongProvider.create(query, key)
            return [provider]
        except ProviderError:
            # No provider is available for the requested key
            # This might be because the platform of that archived song is not available anymore
            # -> treat the request like a search query
            # delete the key so the archived song for the different platform is disregarded
            key = None

    providers: List[MusicProvider] = []
    for platform in enabled_platforms_by_priority():
        module = importlib.import_module(f"core.musiq.{platform}")
        if playlist:
            provider_class = getattr(module, f"{platform.title()}PlaylistProvider")
        else:
            provider_class = getattr(module, f"{platform.title()}SongProvider")
        try:
            provider = provider_class(query, key)
            if platform == preferred_platform:
                providers.insert(0, provider)
            else:
                providers.append(provider)
        except WrongUrlError:
            pass

    if not providers:
        raise ProviderError("No backend configured to handle your request.")

    return providers


def try_providers(session_key: str, providers: List[MusicProvider]) -> MusicProvider:
    """Goes through every given provider and tries to request its music.
    Returns the first provider that was successful with an empty error.
    If unsuccessful, return the last provider."""

    fallback = False
    last_provider = providers[-1]
    provider = providers[0]
    for provider in providers:
        try:
            provider.request(session_key)
            # the current provider could provide the song, don't try the other ones
            break
        except ProviderError:
            # this provider cannot provide this song, use the next provider
            # if this was the last provider, show its error
            # in new music only mode, do not allow fallbacks
            if storage.get("new_music_only") or provider == last_provider:
                return provider
            fallback = True
    provider.error = ""
    if fallback:
        provider.ok_message += " (used fallback)"
    return provider


# accessed by the discord bot
@csrf_exempt
@user_manager.tracked
def request_music(request: WSGIRequest) -> HttpResponse:
    """Endpoint to request music."""
    key_param = request.POST.get("key")
    query = request.POST.get("query")
    playlist = request.POST.get("playlist") == "true"
    platform = request.POST.get("platform")

    if query is None:
        return HttpResponseBadRequest("No query given")
    key = None
    if key_param:
        key = int(key_param)

    try:
        providers = get_providers(query, key, playlist, platform)
    except ProviderError as error:
        return HttpResponseBadRequest(str(error))

    provider = try_providers(request.session.session_key, providers)
    if provider.error:
        return HttpResponseBadRequest(provider.error)

    queue_key = None
    if not playlist:
        assert isinstance(provider, SongProvider)
        queued_song = provider.queued_song
        if not queued_song:
            logging.error(
                "no placeholder was created for query '%s' and key '%s'", query, key
            )
            return HttpResponseBadRequest("No placeholder was created")
        queue_key = queued_song.id

        if storage.get("ip_checking"):
            assert queue_key
            user_manager.try_vote(user_manager.get_client_ip(request), queue_key, 1)

        if storage.get("color_indication") != storage.Privileges.nobody:
            user_manager.register_song(request, queue_key)
            user_manager.register_vote(request, queue_key, 1)

    return JsonResponse({"message": provider.ok_message, "key": queue_key})


@user_manager.tracked
def request_radio(request: WSGIRequest) -> HttpResponse:
    """Endpoint to request radio for the current song."""
    try:
        current_song = CurrentSong.objects.get()
    except CurrentSong.DoesNotExist:
        return HttpResponseBadRequest("Need a song to play the radio")
    provider = SongProvider.create(external_url=current_song.external_url)
    return provider.request_radio(request.session.session_key)


@user_manager.tracked
def index(request: WSGIRequest) -> HttpResponse:
    """Renders the /musiq page."""
    from core import urls

    context = base.context(request)
    context["urls"] = urls.musiq_paths
    context["additional_keywords"] = storage.get("additional_keywords")
    context["forbidden_keywords"] = storage.get("forbidden_keywords")
    context["client_streaming"] = storage.get("output") == "client"
    context["show_stream"] = storage.get("output") in ["client", "icecast"] and (
        not storage.get("privileged_stream") or context["controls_enabled"]
    )
    for platform in ["youtube", "spotify", "soundcloud", "jamendo"]:
        if (
            storage.get("online_suggestions")
            and not storage.get("new_music_only")
            and storage.get(cast(PlatformEnabled, f"{platform}_enabled"))
        ):
            suggestion_count = storage.get(
                cast(PlatformSuggestions, f"{platform}_suggestions")
            )
        else:
            suggestion_count = 0
        context[f"{platform}_suggestions"] = suggestion_count
    return render(request, "musiq.html", context)


def _add_color_indication(engagement, song_dict) -> None:
    requested_by = None
    votes = {}
    if engagement is not None:
        requested_by, votes = ast.literal_eval(engagement)
    song_dict["requestedBy"] = user_manager.color_of(requested_by)
    song_dict["requesterVote"] = votes.get(requested_by, 0)
    song_dict["upvotes"] = []
    song_dict["downvotes"] = []
    for session_key, amount in votes.items():
        if session_key == requested_by:
            continue
        color = user_manager.color_of(session_key)
        if amount > 0:
            song_dict["upvotes"].append(color)
        else:
            song_dict["downvotes"].append(color)


def state_dict() -> Dict[str, Any]:
    """Extends the base state with musiq-specific information and returns it."""
    state = base.state_dict()

    musiq_state: Dict[str, Any] = {}

    musiq_state["paused"] = storage.get("paused")
    musiq_state["shuffle"] = storage.get("shuffle")
    musiq_state["repeat"] = storage.get("repeat")
    musiq_state["autoplay"] = storage.get("autoplay")
    musiq_state["volume"] = storage.get("volume")

    try:
        current_song = CurrentSong.objects.get()
        current_song_dict = model_to_dict(current_song)
        current_song_dict = util.camelize(current_song_dict)
        current_song_dict["durationFormatted"] = song_utils.format_seconds(
            current_song_dict["duration"]
        )
        if storage.get("color_indication") != storage.Privileges.nobody:
            engagement = redis.connection.get(f"engagement-{current_song.queue_key}")
            _add_color_indication(engagement, current_song_dict)
        musiq_state["currentSong"] = current_song_dict

        paused = storage.get("paused")
        if paused:
            progress = (current_song.last_paused - current_song.created).total_seconds()
        else:
            progress = (timezone.now() - current_song.created).total_seconds()
        try:
            progress /= current_song.duration
        except ZeroDivisionError:
            progress = 1
        musiq_state["progress"] = progress * 100
    except CurrentSong.DoesNotExist:
        musiq_state["currentSong"] = None
        musiq_state["paused"] = True
        musiq_state["progress"] = 0

    song_queue = []
    total_time = 0
    all_songs = queue.all()
    if storage.get("interactivity") in [
        storage.Interactivity.upvotes_only,
        storage.Interactivity.full_voting,
    ]:
        all_songs = all_songs.order_by("-votes", "index")
    for song in all_songs:
        song_dict = model_to_dict(song)
        song_dict = util.camelize(song_dict)
        song_dict["durationFormatted"] = song_utils.format_seconds(
            song_dict["duration"]
        )
        if storage.get("color_indication"):
            engagement = redis.connection.get(f"engagement-{song.id}")
            _add_color_indication(engagement, song_dict)
        song_queue.append(song_dict)
        if song_dict["duration"] < 0:
            # skip duration of placeholders
            continue
        total_time += song_dict["duration"]
    musiq_state["totalTimeFormatted"] = song_utils.format_seconds(total_time)
    musiq_state["songQueue"] = song_queue

    if state["alarm"]:
        musiq_state["currentSong"] = {
            "queueKey": -1,
            "manuallyRequested": False,
            "votes": 0,
            "created": "",
            # https://github.com/python/mypy/issues/4976
            **util.camelize(cast(Dict[Any, Any], get_alarm_metadata())),
        }
        musiq_state["progress"] = 0
        musiq_state["paused"] = False
    elif redis.get("backup_playing"):
        musiq_state["currentSong"] = {
            "queueKey": -1,
            "manuallyRequested": False,
            "votes": 0,
            "internalUrl": "backup_stream",
            "externalUrl": storage.get("backup_stream"),
            "artist": "",
            "title": "Backup Stream",
            "duration": 60 * 60 * 24,
            "created": "",
        }
        musiq_state["paused"] = False

    state["musiq"] = musiq_state
    return state


def update_state() -> None:
    """Sends an update event to all connected clients."""
    send_state(state_dict())



"""Contains the core app configuration."""
import atexit
import logging
import os
import sys

from django.apps import AppConfig

from django.conf import settings as conf
from core.util import strtobool


class CoreConfig(AppConfig):
    """This is the configuration for the core app."""

    name = "core"

    def ready(self) -> None:
        if "celery" in sys.argv and strtobool(os.environ.get("RUN_MAIN", "0")):
            # if the development celery process starts,
            # have it import all modules containing celery tasks
            # this way, its autoreload is notified on changes in these modules

            for module in conf.CELERY_IMPORTS:
                assert isinstance(module, str)
                __import__(module)

            return

        # ready is called for every management command and for autoreload
        # only start raveberry when
        # in debug mode and the main application is run (not autoreload)
        # or in prod mode (run by daphne)
        start_raveberry = (
            strtobool(os.environ.get("RUN_MAIN", "0"))
            if "runserver" in sys.argv
            else sys.argv[0].endswith("daphne")
        )

        if conf.TESTING:
            from core.musiq import controller

            # when MopidyAPI instances are created too often,
            # mopidy runs into a "Set changed size during iteration" error.
            # Thus we initialize the interface once and not for every testcase
            controller.start()

        if start_raveberry:
            from core import tasks
            from core import redis
            from core.musiq import musiq
            from core.musiq import playback
            from core.settings import basic
            from core.settings import platforms
            from core.lights import worker

            logging.info("starting raveberry")

            redis.start()
            tasks.start()

            worker.start()
            # platforms needs to start before musiq because mopidy_available is checked there
            platforms.start()
            musiq.start()
            basic.start()

            def stop_workers() -> None:
                # wake up the playback thread and stop it
                redis.put("stop_playback_loop", True)
                playback.queue_changed.set()

                # wake the buzzer thread so it exits
                playback.buzzer_stopped.set()

                # wake up the listener thread with an instruction to stop the lights worker
                redis.connection.publish("lights_settings_changed", "stop")

            atexit.register(stop_workers)



"""This module contains the rendertemplate command."""
import json

from bs4 import BeautifulSoup
from django.core.management.base import BaseCommand, CommandError
from django.template.loader import TemplateDoesNotExist, get_template


class Command(BaseCommand):
    """rendertemplate renders the given template with the given context. Used in testing."""

    help = "Renders the specified template writes the head and body tag to the given output files"

    def add_arguments(self, parser):
        parser.add_argument("template")
        parser.add_argument("head_file")
        parser.add_argument("body_file")
        parser.add_argument("context", nargs="?")

    def handle(self, *args, **options):
        try:
            template = get_template(options["template"])
        except TemplateDoesNotExist as error:
            raise CommandError(
                f"Template \"{options['template']}\" does not exist"
            ) from error

        if options["context"]:
            context = json.loads(options["context"])
        else:
            context = {}
        html = template.render(context)

        soup = BeautifulSoup(html, "html.parser")
        with open(options["head_file"], "w", encoding="utf-8") as head_file:
            head_file.write(soup.head.prettify())
        with open(options["body_file"], "w", encoding="utf-8") as body_file:
            body_file.write(soup.body.prettify())



"""This module handles all controls that influence visualization in general."""

from __future__ import annotations

import socket
from functools import wraps
from typing import cast, Tuple, Callable

from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse, HttpResponseForbidden, HttpResponseBadRequest

from core import user_manager, redis
from core.lights import lights
from core.settings import storage
from core.settings.storage import DeviceBrightness, DeviceMonochrome, DeviceProgram
from core.util import extract_value, strtobool


def control(func: Callable) -> Callable:
    """A decorator for functions that control the lights.
    Every control changes the views state and returns an empty response.
    Mod privileges are required to control the lights."""

    def _decorator(request: WSGIRequest) -> HttpResponse:
        if not user_manager.has_controls(request.user):
            return HttpResponseForbidden()

        response = func(request)
        lights.update_state()
        if response is not None:
            return response
        return HttpResponse()

    return wraps(func)(_decorator)


def _notify_settings_changed(settings: str) -> None:
    # notifies the worker thread that the given settings (eg <device> or "base") changed
    redis.connection.publish("lights_settings_changed", settings)


def alarm_started() -> None:
    """Notifies the lights worker that the alarm was started."""
    _notify_settings_changed("alarm_started")


def alarm_stopped() -> None:
    """Notifies the lights worker that the alarm was stopped."""
    _notify_settings_changed("alarm_stopped")


def persist_program_change(device, program) -> None:
    """Persist the program in the database."""
    assert device in ["ring", "strip", "wled", "screen"]
    current_program = storage.get(cast(DeviceProgram, f"{device}_program"))
    storage.put(cast(DeviceProgram, f"last_{device}_program"), current_program)
    storage.put(cast(DeviceProgram, f"{device}_program"), program)


def set_program(device: str, program: str) -> None:
    """Updates the given device to the given program."""
    persist_program_change(device, program)
    _notify_settings_changed(device)


@control
def set_lights_shortcut(request: WSGIRequest) -> HttpResponse:
    """Stores the current lights state and restores the previous one."""
    value, response = extract_value(request.POST)
    should_enable = strtobool(value)
    is_enabled = (
        storage.get("ring_program") != "Disabled"
        or storage.get("wled_program") != "Disabled"
        or storage.get("strip_program") != "Disabled"
    )
    if should_enable == is_enabled:
        return HttpResponse()
    if should_enable:
        for device in ["ring", "wled", "strip"]:
            set_program(
                device, storage.get(cast(DeviceProgram, f"last_{device}_program"))
            )
    else:
        for device in ["ring", "wled", "strip"]:
            set_program(device, "Disabled")
    return response


@control
def set_ups(request: WSGIRequest) -> HttpResponse:
    """Updates the global speed of programs supporting it."""
    value, response = extract_value(request.POST)
    storage.put("ups", float(value))
    _notify_settings_changed("base")
    return response


@control
def set_program_speed(request: WSGIRequest) -> HttpResponse:
    """Updates the global speed of programs supporting it."""
    value, response = extract_value(request.POST)
    storage.put("program_speed", float(value))
    _notify_settings_changed("base")
    return response


@control
def set_fixed_color(request: WSGIRequest) -> HttpResponse:
    """Updates the static color used for some programs."""
    hex_col, response = extract_value(request.POST)
    hex_col = hex_col.lstrip("#")
    try:
        color = tuple(int(hex_col[i : i + 2], 16) / 255 for i in (0, 2, 4))
    except IndexError:
        return HttpResponseBadRequest("color needs to be in #rrggbb format")
    # https://github.com/python/mypy/issues/5068
    color = cast(Tuple[float, float, float], color)
    storage.put("fixed_color", color)
    _notify_settings_changed("base")
    return response


def _handle_program_request(device: str, request: WSGIRequest) -> HttpResponse:
    program, response = extract_value(request.POST)
    assert device in ["ring", "strip", "wled", "screen"]
    if program == storage.get(cast(DeviceProgram, f"{device}_program")):
        # the program doesn't change, return immediately
        return HttpResponse()
    set_program(device, program)
    return response


def _handle_brightness_request(device: str, request: WSGIRequest) -> HttpResponse:
    value, response = extract_value(request.POST)
    assert device in ["ring", "strip", "wled", "screen"]
    storage.put(cast(DeviceBrightness, f"{device}_brightness"), float(value))
    _notify_settings_changed(device)
    return response


def _handle_monochrome_request(device: str, request: WSGIRequest) -> HttpResponse:
    value, response = extract_value(request.POST)
    assert device in ["ring", "strip", "wled", "screen"]
    storage.put(cast(DeviceMonochrome, f"{device}_monochrome"), strtobool(value))
    _notify_settings_changed(device)
    return response


@control
def set_ring_program(request: WSGIRequest) -> HttpResponse:
    """Updates the ring program."""
    return _handle_program_request("ring", request)


@control
def set_ring_brightness(request: WSGIRequest) -> HttpResponse:
    """Updates the ring brightness."""
    return _handle_brightness_request("ring", request)


@control
def set_ring_monochrome(request: WSGIRequest) -> HttpResponse:
    """Sets whether the ring should be in one color only."""
    return _handle_monochrome_request("ring", request)


@control
def set_wled_led_count(request: WSGIRequest) -> HttpResponse:
    """Updates the wled led_count."""
    value, response = extract_value(request.POST)
    if not 2 <= int(value) <= 490:
        return HttpResponseBadRequest("must be between 2 and 490")
    storage.put("wled_led_count", int(value))
    _notify_settings_changed("wled")
    return response


@control
def set_wled_ip(request: WSGIRequest) -> HttpResponse:
    """Updates the wled ip."""
    value, response = extract_value(request.POST)
    try:
        socket.inet_aton(value)
    except socket.error:
        return HttpResponseBadRequest("invalid ip")
    storage.put("wled_ip", value)
    _notify_settings_changed("wled")
    return response


@control
def set_wled_port(request: WSGIRequest) -> HttpResponse:
    """Updates the wled port."""
    value, response = extract_value(request.POST)
    if not 1024 <= int(value) <= 65535:
        return HttpResponseBadRequest("invalid port")
    storage.put("wled_port", int(value))
    _notify_settings_changed("wled")
    return response


@control
def set_wled_program(request: WSGIRequest) -> HttpResponse:
    """Updates the wled program."""
    return _handle_program_request("wled", request)


@control
def set_wled_brightness(request: WSGIRequest) -> HttpResponse:
    """Updates the wled brightness."""
    return _handle_brightness_request("wled", request)


@control
def set_wled_monochrome(request: WSGIRequest) -> HttpResponse:
    """Sets whether the wled should be in one color only."""
    return _handle_monochrome_request("wled", request)


@control
def set_strip_program(request: WSGIRequest) -> HttpResponse:
    """Updates the strip program."""
    return _handle_program_request("strip", request)


@control
def set_strip_brightness(request: WSGIRequest) -> HttpResponse:
    """Updates the strip brightness."""
    return _handle_brightness_request("strip", request)


@control
def adjust_screen(_request: WSGIRequest) -> None:
    """Adjusts the resolution of the screen."""
    _notify_settings_changed("adjust_screen")


@control
def set_screen_program(request: WSGIRequest) -> HttpResponse:
    """Updates the screen program."""
    return _handle_program_request("screen", request)


@control
def set_initial_resolution(request: WSGIRequest) -> HttpResponse:
    """Sets the resolution used on the screen."""
    value, response = extract_value(request.POST)
    resolution = tuple(map(int, value.split("x")))
    resolution = cast(Tuple[int, int], resolution)
    storage.put("initial_resolution", resolution)
    # adjusting sets the resolution and restarts the screen program
    _notify_settings_changed("adjust_screen")
    return response


@control
def set_dynamic_resolution(request: WSGIRequest) -> HttpResponse:
    """Sets whether the resolution should be dynamically adjusted depending on the performance."""
    value, response = extract_value(request.POST)
    storage.put("dynamic_resolution", strtobool(value))
    _notify_settings_changed("base")
    return response



"""This module defines the abstract interface to interface with playback software."""

from django.conf import settings as conf
from mopidyapi.client import MopidyAPI

from core import redis
from core.musiq import musiq
from core.settings import settings


def set_playback_error(error: bool) -> None:
    """Sets the playback error. Updates musiq state when the state changes."""
    if redis.get("playback_error"):
        if not error:
            redis.put("playback_error", False)
            musiq.update_state()
            settings.update_state()
    else:
        if error:
            redis.put("playback_error", True)
            musiq.update_state()
            settings.update_state()


class Player:
    """Class containing methods to interface with playback software."""

    def start_song(self, song, catch_up: float):
        """Start's the specified song.
        Raises a PlaybackError on error."""
        raise NotImplementedError

    def should_stop_waiting(self, previous_error: bool) -> bool:
        """check whether the main loop should stop waiting for the song to end.
        Returns True if it should stop waiting, False otherwise.
        Raises PlaybackError on error."""
        return False

    def play_alarm(self, interrupt: bool, alarm_path: str) -> None:
        raise NotImplementedError

    def play_backup_stream(self):
        raise NotImplementedError


def _active_player():
    """Looks up the currently active player and returns the respective class.
    This is necessary since the requests have no reference to a player object,
    so all methods need to be static.
    Thus no inheritance is possible, and needs to be faked here."""

    active_player = redis.get("active_player")
    if active_player == "mopidy":
        from core.musiq import mopidy_player

        return mopidy_player
    elif active_player == "spotify":
        from core.musiq import spotify_player

        return spotify_player
    elif active_player == "fake":
        from core.musiq import fake_player

        return fake_player


def restart() -> None:
    """Restarts the current song from the beginning."""
    _active_player().restart()


def seek_backward(seek_distance: float) -> None:
    """Jumps back in the current song."""
    _active_player().seek_backward(seek_distance)


def play() -> None:
    """Resumes the current song if it is paused.
    No-op if already playing."""
    _active_player().play()


def pause() -> None:
    """Pauses the current song if it is playing.
    No-op if already paused."""
    _active_player().pause()


def seek_forward(seek_distance: float) -> None:
    """Jumps forward in the current song."""
    _active_player().seek_forward(seek_distance)


def skip() -> None:
    """Skips the current song and continues with the next one."""
    _active_player().skip()


def set_volume(volume) -> None:
    """Sets the playback volume.
    value has to be a float between 0 and 1."""
    _active_player().set_volume(volume)



"""This module handles all settings regarding the music platforms."""
from __future__ import annotations

import importlib.util
import os
import subprocess
from typing import cast

from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse, HttpResponseBadRequest

from django.conf import settings as conf
from core import redis
from core.settings import library, storage, system
from core.settings.settings import control
from core.settings.storage import PlatformEnabled
from core.util import extract_value, strtobool


def start() -> None:
    """Initializes this module by checking which platforms are available to use."""

    # mopidy is available if the server can be connected to
    import socket

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    location = (conf.MOPIDY_HOST, int(conf.MOPIDY_PORT))
    mopidy_available = True
    try:
        check = sock.connect_ex(location)
        if check != 0:
            mopidy_available = False
    except (ConnectionRefusedError, socket.gaierror):
        # check for gaierror because docker compose setups might not be able to resolve the hostname
        mopidy_available = False
    finally:
        sock.close()
    redis.put("mopidy_available", mopidy_available)

    # local songs are enabled if a library is set
    local_enabled = os.path.islink(library.get_library_path())
    storage.put("local_enabled", local_enabled)

    # in the docker container all dependencies are installed
    youtube_available = conf.DOCKER or importlib.util.find_spec("yt_dlp") is not None
    redis.put("youtube_available", youtube_available)
    if not youtube_available:
        # if youtube is not available, overwrite the database to disable it
        storage.put("youtube_enabled", False)

    spotify_available = conf.DOCKER or importlib.util.find_spec("spotipy") is not None
    redis.put("spotify_available", spotify_available)
    if not spotify_available:
        storage.put("spotify_enabled", False)

    soundcloud_available = (
        conf.DOCKER or importlib.util.find_spec("soundcloud") is not None
    )
    redis.put("soundcloud_available", soundcloud_available)
    if not soundcloud_available:
        storage.put("soundcloud_enabled", False)

    # Jamendo has no python dependencies we could easily check.
    try:
        jamendo_available = (
            conf.DOCKER
            or "[jamendo]"
            in subprocess.check_output(["mopidy", "config"], stderr=subprocess.DEVNULL)
            .decode()
            .splitlines()
        )
    except FileNotFoundError:
        jamendo_available = True
    redis.put("jamendo_available", jamendo_available)
    if not jamendo_available:
        storage.put("jamendo_enabled", False)


@control
def set_youtube_enabled(request: WSGIRequest) -> HttpResponse:
    """Enables or disables youtube to be used as a song provider."""
    value, response = extract_value(request.POST)
    storage.put("youtube_enabled", strtobool(value))
    return response


@control
def set_youtube_suggestions(request: WSGIRequest) -> HttpResponse:
    """Sets the number of online suggestions from youtube to be shown."""
    value, response = extract_value(request.POST)
    storage.put("youtube_suggestions", int(value))
    return response


def _set_extension_enabled(extension, enabled) -> HttpResponse:
    if enabled:
        if conf.DOCKER:
            response = HttpResponse(
                "Make sure you provided mopidy with correct credentials."
            )
        else:
            extensions = system.check_mopidy_extensions()
            functional, message = extensions[extension]
            if not functional:
                return HttpResponseBadRequest(message)
            response = HttpResponse(message)
    else:
        response = HttpResponse("Disabled extension")
    assert extension in ["local", "youtube", "spotify", "soundcloud", "jamendo"]
    storage.put(cast(PlatformEnabled, f"{extension}_enabled"), enabled)
    return response


@control
def set_spotify_enabled(request: WSGIRequest) -> HttpResponse:
    """Enables or disables spotify to be used as a song provider.
    Makes sure mopidy has correct spotify configuration."""
    value, _ = extract_value(request.POST)
    mopidy_response = _set_extension_enabled("spotify", strtobool(value))

    if not strtobool(value):
        return mopidy_response
    else:
        # Check whether Spotify device authorization works
        from core.musiq.spotify import Spotify
        from spotipy.oauth2 import SpotifyOauthError

        try:
            api = Spotify.device_api()
            api.devices()
        except SpotifyOauthError as e:
            local_message = e.error_description or e
            local_successful = False
        else:
            local_message = "Login successful"
            local_successful = True
        message = f"device: {local_message}; local: {mopidy_response.content.decode()}"
        if type(mopidy_response) is HttpResponseBadRequest and not local_successful:
            return HttpResponseBadRequest(message)
        else:
            # enable spotify if at least one option works
            storage.put(cast(PlatformEnabled, f"spotify_enabled"), True)
            return HttpResponse(message)


@control
def set_spotify_suggestions(request: WSGIRequest) -> HttpResponse:
    """Sets the number of online suggestions from spotify to be shown."""
    value, response = extract_value(request.POST)
    storage.put("spotify_suggestions", int(value))
    return response


@control
def set_spotify_mopidy_credentials(request: WSGIRequest) -> HttpResponse:
    """Update spotify mopidy credentials."""
    username = request.POST.get("username")
    password = request.POST.get("password")
    client_id = request.POST.get("client_id")
    client_secret = request.POST.get("client_secret")

    if not username or not password or not client_id or not client_secret:
        return HttpResponseBadRequest("All fields are required")

    storage.put("spotify_username", username)
    storage.put("spotify_password", password)
    storage.put("spotify_mopidy_client_id", client_id)
    storage.put("spotify_mopidy_client_secret", client_secret)

    system.update_mopidy_config("pulse")
    from core.musiq.spotify import Spotify

    Spotify.create_mopidy_api()
    return HttpResponse("Updated credentials")


@control
def set_spotify_device_credentials(request: WSGIRequest) -> HttpResponse:
    """Update spotify device credentials."""
    client_id = request.POST.get("client_id")
    client_secret = request.POST.get("client_secret")
    redirect_uri = request.POST.get("redirect_uri")
    authorized_url = request.POST.get("authorized_url")

    if not client_id or not client_secret or not redirect_uri or not authorized_url:
        return HttpResponseBadRequest("All fields are required")

    storage.put("spotify_device_client_id", client_id)
    storage.put("spotify_device_client_secret", client_secret)
    storage.put("spotify_redirect_uri", redirect_uri)
    storage.put("spotify_authorized_url", authorized_url)

    system.update_mopidy_config("fakesink")
    from core.musiq.spotify import Spotify

    Spotify.create_device_api()
    return HttpResponse("Updated credentials")


@control
def set_soundcloud_enabled(request: WSGIRequest) -> HttpResponse:
    """Enables or disables soundcloud to be used as a song provider.
    Makes sure mopidy has correct soundcloud configuration."""
    value, _ = extract_value(request.POST)
    return _set_extension_enabled("soundcloud", strtobool(value))


@control
def set_soundcloud_suggestions(request: WSGIRequest) -> HttpResponse:
    """Sets the number of online suggestions from soundcloud to be shown."""
    value, response = extract_value(request.POST)
    storage.put("soundcloud_suggestions", int(value))
    return response


@control
def set_soundcloud_credentials(request: WSGIRequest) -> HttpResponse:
    """Update soundcloud credentials."""
    auth_token = request.POST.get("auth_token")

    if not auth_token:
        return HttpResponseBadRequest("All fields are required")

    storage.put("soundcloud_auth_token", auth_token)

    system.update_mopidy_config("pulse")
    return HttpResponse("Updated credentials")


@control
def set_jamendo_enabled(request: WSGIRequest) -> HttpResponse:
    """Enables or disables jamendo to be used as a song provider.
    Makes sure mopidy has correct jamendo configuration."""
    value, _ = extract_value(request.POST)
    return _set_extension_enabled("jamendo", strtobool(value))


@control
def set_jamendo_suggestions(request: WSGIRequest) -> HttpResponse:
    """Sets the number of online suggestions from jamendo to be shown."""
    value, response = extract_value(request.POST)
    storage.put("jamendo_suggestions", int(value))
    return response


@control
def set_jamendo_credentials(request: WSGIRequest) -> HttpResponse:
    """Update jamendo credentials."""
    client_id = request.POST.get("client_id")

    if not client_id:
        return HttpResponseBadRequest("All fields are required")

    storage.put("jamendo_client_id", client_id)

    system.update_mopidy_config("pulse")
    return HttpResponse("Updated credentials")



"""This module contains all programs used for visualization (except complex screen programs)."""

from __future__ import annotations

import errno
import logging
import os
import subprocess
from typing import Tuple, List, Optional, TYPE_CHECKING

from django.conf import settings as conf

from core.lights import leds

if TYPE_CHECKING:
    from core.lights.worker import DeviceManager


class LightProgram:
    """The base class for all programs."""

    def __init__(self, manager: "DeviceManager", name: str) -> None:
        self.manager = manager
        self.consumers = 0
        self.name = name

    def start(self) -> None:
        """Initializes the program, allocates resources."""

    def use(self) -> None:
        """Tells the program that it is used by another consumer.
        Starts the program if this is the first usage."""
        if self.consumers == 0:
            self.start()
        self.consumers += 1

    def compute(self) -> None:
        """Is called once per update. Any computation should happen here,
        so it can be reused in the returning functions"""

    def stop(self) -> None:
        """Stops the program, releases resources."""

    def release(self) -> None:
        """Tells the program that one consumer does not use it anymore.
        Stops the program if this was the last one."""
        self.consumers -= 1
        if self.consumers == 0:
            self.stop()


class LedProgram(LightProgram):
    """The base class for all led visualization programs."""

    def ring_colors(self) -> List[Tuple[float, float, float]]:
        """Returns the colors for the ring, one rgb tuple for each led."""
        raise NotImplementedError()

    def wled_colors(self) -> List[Tuple[float, float, float]]:
        """Returns the colors for WLED, one rgb tuple for each led."""
        raise NotImplementedError()

    def strip_color(self) -> Tuple[float, float, float]:
        """Returns the rgb values for the strip."""
        raise NotImplementedError()


class ScreenProgram(LightProgram):
    """Base calls for all programs showing something on screen."""


class Disabled(LedProgram, ScreenProgram):
    """A null class to represent inactivity."""

    def __init__(self, manager: "DeviceManager") -> None:
        super().__init__(manager, "Disabled")

    def ring_colors(self) -> List[Tuple[float, float, float]]:
        raise NotImplementedError()

    def wled_colors(self) -> List[Tuple[float, float, float]]:
        raise NotImplementedError()

    def strip_color(self) -> Tuple[float, float, float]:
        raise NotImplementedError()


class Alarm(LightProgram):
    """This program makes the leds flash red in sync to the played sound.
    Only computes the brightness, does not display it."""

    INCREASING_DURATION = 0.45
    DECREASING_DURATION = 0.8
    SOUND_DURATION = 2.1
    SOUND_REPETITION = 2.5

    def __init__(self, manager: "DeviceManager") -> None:
        super().__init__(manager, "Alarm")
        self.time_passed = 0.0
        self.sound_count = 0
        # only during this program, thus False by default
        self.pwr_led_enabled = False
        self.factor = -1.0

    def start(self) -> None:
        self.time_passed = 0.0
        self.sound_count = 0
        self.factor = 0

    def compute(self) -> None:
        """If active, compute the brightness for the red color,
        depending on the time that has passed since starting the sound."""
        # do not compute if the alarm is not active
        if self.consumers == 0:
            return
        self.time_passed += 1 / self.manager.settings["ups"]
        if self.time_passed >= Alarm.SOUND_REPETITION:
            self.sound_count += 1
            self.time_passed %= Alarm.SOUND_REPETITION

        if self.sound_count >= 4:
            self.factor = 0
            return
        if self.time_passed < Alarm.INCREASING_DURATION:
            self.factor = self.time_passed / Alarm.INCREASING_DURATION
        elif self.time_passed < Alarm.SOUND_DURATION - Alarm.DECREASING_DURATION:
            self.factor = 1
        elif self.time_passed < Alarm.SOUND_DURATION:
            self.factor = (
                1
                - (
                    self.time_passed
                    - (Alarm.SOUND_DURATION - Alarm.DECREASING_DURATION)
                )
                / Alarm.DECREASING_DURATION
            )
        else:
            self.factor = 0

        # Ideally, the pwr led would flash with increasing frequency,
        # but the Pi can not handle enough script executions to make it look good.
        if self.pwr_led_enabled and self.factor < 0.7:
            leds.disable_pwr_led()
            self.pwr_led_enabled = False
        elif not self.pwr_led_enabled and self.factor >= 0.7:
            leds.enable_pwr_led()
            self.pwr_led_enabled = True

    def stop(self) -> None:
        self.factor = -1.0


class Cava(LightProgram):
    """This Program manages the interaction with cava.
    It provides the current frequencies for other programs to use."""

    # Keep these configurations in sync with config/cava.config
    BARS = 256
    BIT_FORMAT = 8

    def __init__(self, manager: "DeviceManager") -> None:
        super().__init__(manager, "Cava")

        self.cava_fifo_path = os.path.join(conf.BASE_DIR, "config/cava_fifo")

        self.frame_length = Cava.BARS * (Cava.BIT_FORMAT // 8)

        self.current_frame: List[float] = []
        self.growing_frame = b""
        self.cava_process: Optional[subprocess.Popen] = None
        self.cava_fifo = -1

    def start(self) -> None:
        self.current_frame = [0 for _ in range(Cava.BARS)]
        self.growing_frame = b""
        try:
            # delete old contents of the pipe
            os.remove(self.cava_fifo_path)
        except FileNotFoundError:
            # the file does not exist
            pass
        try:
            old_umask = os.umask(0o002)
            os.mkfifo(self.cava_fifo_path)
        except FileExistsError:
            # the file already exists
            logging.info("%s already exists while starting", self.cava_fifo_path)
        finally:
            os.umask(old_umask)

        self.manager.set_cava_framerate()
        self.cava_process = subprocess.Popen(
            ["cava", "-p", os.path.join(conf.BASE_DIR, "config/cava.config")],
            cwd=conf.BASE_DIR,
            env={"PULSE_SERVER": conf.PULSE_SERVER, **os.environ},
        )
        self.cava_fifo = os.open(self.cava_fifo_path, os.O_RDONLY | os.O_NONBLOCK)

    def compute(self) -> None:
        """If active, read output from the cava program.
        Make sure that the most recent frame is always fully available,
        Stores incomplete frames for the next update."""
        # do not compute if no program uses cava
        if self.consumers == 0:
            return
        # read the fifo until we get to the current frame
        while True:
            try:
                read = os.read(
                    self.cava_fifo, self.frame_length - len(self.growing_frame)
                )
                if read == b"":
                    return
                self.growing_frame += read
            except OSError as error:
                if error.errno in (errno.EAGAIN, errno.EWOULDBLOCK):
                    # there were not enough bytes for a whole frame, keep the old frame
                    return

            # we read a whole frame, update the factors
            if len(self.growing_frame) == self.frame_length:
                self.current_frame = [int(b) / 255 for b in self.growing_frame]
                self.growing_frame = b""

    def stop(self) -> None:
        try:
            os.close(self.cava_fifo)
        except OSError as error:
            logging.info("fifo already closed: %s", error)
        except TypeError as error:
            logging.info("fifo does not exist: %s", error)

        if self.cava_process:
            self.cava_process.terminate()

        try:
            os.remove(self.cava_fifo_path)
        except FileNotFoundError as error:
            # the file was already deleted
            logging.info("%s not found while deleting: %s", self.cava_fifo_path, error)



"""This module handles the suggestions when starting to
type into the input field on the musiq page."""

from __future__ import annotations

import random
import threading
from typing import Any, Dict, Iterable, List, Mapping, TypedDict, Union, cast

from django.core.handlers.wsgi import WSGIRequest
from django.db.models import F, Q
from django.db.models.functions import Greatest
from django.http import HttpResponseBadRequest
from django.http.response import HttpResponse, JsonResponse

from core import redis
from core.models import ArchivedPlaylist, ArchivedQuery, ArchivedSong
from core.musiq import song_utils
from core.settings import storage
from core.settings.storage import PlatformEnabled, PlatformSuggestions
from main import settings

u_values_list = [
    "u_id",
    "u_url",
    "u_artist",
    "u_title",
    "u_duration",
    "u_counter",
    "u_cached",
]


class SuggestionResult(TypedDict, total=False):
    """Type that encapsulates a suggestion for a query."""

    key: Union[int, str]
    value: str
    type: str
    counter: int  # optional, not in online suggestions
    durationFormatted: str  # optional, not in online or playlist suggestions
    confusable: bool  # optional, only in song results


def random_suggestion(request: WSGIRequest) -> HttpResponse:
    """This method returns a random suggestion from the database.
    Depending on the value of :param playlist:,
    either a previously pushed playlist or song is returned."""
    suggest_playlist = request.GET["playlist"] == "true"
    if not suggest_playlist:
        if ArchivedSong.objects.count() == 0:
            return HttpResponseBadRequest("No songs to suggest from")
        index = random.randint(0, ArchivedSong.objects.count() - 1)
        song = ArchivedSong.objects.all()[index]
        return JsonResponse({"suggestion": song.displayname(), "key": song.id})

    # exclude radios from suggestions
    remaining_playlists = (
        ArchivedPlaylist.objects.all()
        .exclude(list_id__startswith="RD")
        .exclude(list_id__contains="&list=RD")
    )
    if remaining_playlists.count() == 0:
        return HttpResponseBadRequest("No playlists to suggest from")
    index = random.randint(0, remaining_playlists.count() - 1)
    playlist = remaining_playlists.all()[index]
    return JsonResponse({"suggestion": playlist.title, "key": playlist.id})


def online_suggestions(request: WSGIRequest) -> JsonResponse:
    """Returns online suggestions for a given query."""
    query = request.GET["term"]
    suggest_playlist = request.GET["playlist"] == "true"

    if storage.get("new_music_only"):
        return JsonResponse([], safe=False)

    results: List[SuggestionResult] = []
    if storage.get("online_suggestions") and redis.get("has_internet"):
        threads = []
        results_lock = threading.Lock()

        def fetch_youtube() -> None:
            from core.musiq.youtube import Youtube

            youtube_suggestions = Youtube().get_search_suggestions(query)
            youtube_suggestions = youtube_suggestions[
                : storage.get("youtube_suggestions")
            ]
            with results_lock:
                for suggestion in youtube_suggestions:
                    results.append(
                        {"key": -1, "value": suggestion, "type": "youtube-online"}
                    )

        def fetch_spotify() -> None:
            from core.musiq.spotify import Spotify

            spotify_suggestions = Spotify().get_search_suggestions(
                query, suggest_playlist
            )
            spotify_suggestions = spotify_suggestions[
                : storage.get("spotify_suggestions")
            ]
            with results_lock:
                for suggestion, external_url in spotify_suggestions:
                    results.append(
                        {
                            "key": external_url,
                            "value": suggestion,
                            "type": "spotify-online",
                        }
                    )

        def fetch_soundcloud() -> None:
            from core.musiq.soundcloud import Soundcloud

            soundcloud_suggestions = Soundcloud().get_search_suggestions(query)
            soundcloud_suggestions = soundcloud_suggestions[
                : storage.get("soundcloud_suggestions")
            ]
            with results_lock:
                for suggestion in soundcloud_suggestions:
                    results.append(
                        {"key": -1, "value": suggestion, "type": "soundcloud-online"}
                    )

        def fetch_jamendo() -> None:
            from core.musiq.jamendo import Jamendo

            jamendo_suggestions = Jamendo().get_search_suggestions(query)
            jamendo_suggestions = jamendo_suggestions[
                : storage.get("jamendo_suggestions")
            ]
            with results_lock:
                for suggestion in jamendo_suggestions:
                    results.append(
                        {"key": -1, "value": suggestion, "type": "jamendo-online"}
                    )

        suggestion_fetchers = {
            "youtube": fetch_youtube,
            "spotify": fetch_spotify,
            "soundcloud": fetch_soundcloud,
            "jamendo": fetch_jamendo,
        }

        for platform in ["youtube", "spotify", "soundcloud", "jamendo"]:
            if (
                storage.get(cast(PlatformEnabled, f"{platform}_enabled"))
                and storage.get(cast(PlatformSuggestions, f"{platform}_suggestions"))
                > 0
            ):
                thread = threading.Thread(target=suggestion_fetchers[platform])
                threads.append(thread)
                thread.start()

        for thread in threads:
            thread.join()

    return JsonResponse(results, safe=False)


def _offline_playlist_suggestions(query: str) -> List[SuggestionResult]:
    results: List[SuggestionResult] = []
    terms = query.split()
    remaining_playlists = ArchivedPlaylist.objects.prefetch_related("queries")
    # exclude radios from suggestions
    remaining_playlists = remaining_playlists.exclude(list_id__startswith="RD").exclude(
        list_id__contains="&list=RD"
    )

    # we could have more strict types with TypedDicts,
    # but massaging both QuerySets into a single Type would require asserts/casts,
    # which have little performance impact but should be avoided in this critical codepath
    playlist_results: Iterable[Mapping[str, Any]]
    if settings.DEBUG:
        matching_playlists = remaining_playlists
        for term in terms:
            matching_playlists = matching_playlists.filter(
                Q(title__icontains=term) | Q(queries__query__icontains=term)
            )

        playlist_results = (
            matching_playlists.order_by("-counter")
            .values("id", "title", "counter")
            .distinct()[: storage.get("number_of_suggestions")]
        )
    else:
        from django.contrib.postgres.search import TrigramWordSimilarity

        similar_playlists = remaining_playlists.annotate(
            title_similarity=TrigramWordSimilarity(query, "title"),
            query_similarity=TrigramWordSimilarity(query, "queries__query"),
            max_similarity=Greatest("title_similarity", "query_similarity"),
        )

        playlist_results = (
            similar_playlists.order_by("-max_similarity")
            .values("id", "title", "counter")
            .distinct()[: storage.get("number_of_suggestions")]
        )
    for playlist in playlist_results:
        archived_playlist = ArchivedPlaylist.objects.get(id=playlist["id"])
        platform = song_utils.determine_playlist_type(archived_playlist)
        result_dict: SuggestionResult = {
            "key": playlist["id"],
            "value": playlist["title"],
            "counter": playlist["counter"],
            "type": platform,
        }
        first_entry = archived_playlist.entries.first()
        if first_entry and platform == "local":
            # don't suggest local playlists if their first song is not cached
            # i.e. not at the expected location
            try:
                first_song = ArchivedSong.objects.filter(url=first_entry.url).get()
                if not first_song.cached:
                    continue
            except ArchivedSong.DoesNotExist:
                continue
        else:
            # don't suggest songs if the respective platform is disabled
            assert platform in ["youtube", "spotify", "soundcloud", "jamendo"]
            if not storage.get(cast(PlatformEnabled, f"{platform}_enabled")):
                continue
        results.append(result_dict)
    return results


def _postgres_song_results(query: str) -> List[Dict[str, Union[int, str, float]]]:
    from django.contrib.postgres.search import TrigramWordSimilarity

    # Songs and queries that lead to the songs should both be searched.
    # However, they are stored in two different tables.
    # We first filter for similarity in each table separately (query <> artist, title).
    # Then, we union the subqueries and order the result.

    # We calculate the similarity for the respective other table as well.
    # (query for ArchivedSong and artist,title for ArchivedQuery).
    # This can't happen after the union, union-querysets are limited in functionality.
    # https://docs.djangoproject.com/en/3.2/ref/models/querysets/#union

    # In the select statement of a query,
    # django first lists all values and then all aliases.
    # This is problematic during a union which only cares about the order of the values.
    # To make sure that the right values are combined, annotate all values with an alias.
    # https://stackoverflow.com/questions/60562759/incorrect-results-with-annotate-values-union-in-django

    # The Greatest function of postgres returns the greatest non-null value.
    # For other backends, the similarity values would need to be coalesced to 0
    # otherwise the max_similarity would be null if one of the fields is null

    similar_queries = (
        ArchivedQuery.objects.filter(Q(query__trigram_word_similar=query))
        .annotate(u_id=F("song__id"))
        .annotate(u_url=F("song__url"))
        .annotate(u_artist=F("song__artist"))
        .annotate(u_title=F("song__title"))
        .annotate(u_duration=F("song__duration"))
        .annotate(u_counter=F("song__counter"))
        .annotate(u_cached=F("song__cached"))
        .annotate(u_query=F("query"))
        .annotate(artist_similarity=TrigramWordSimilarity(query, "u_artist"))
        .annotate(title_similarity=TrigramWordSimilarity(query, "u_title"))
        .annotate(query_similarity=TrigramWordSimilarity(query, "u_query"))
        .annotate(
            max_similarity=Greatest(
                "artist_similarity", "title_similarity", "query_similarity"
            )
        )
        .values(*u_values_list, "u_query", "max_similarity")
    )

    similar_songs = (
        ArchivedSong.objects.filter(
            Q(artist__trigram_word_similar=query) | Q(title__trigram_word_similar=query)
        )
        .annotate(u_id=F("id"))
        .annotate(u_url=F("url"))
        .annotate(u_artist=F("artist"))
        .annotate(u_title=F("title"))
        .annotate(u_duration=F("duration"))
        .annotate(u_counter=F("counter"))
        .annotate(u_cached=F("cached"))
        .annotate(u_query=F("queries__query"))
        .annotate(artist_similarity=TrigramWordSimilarity(query, "u_artist"))
        .annotate(title_similarity=TrigramWordSimilarity(query, "u_title"))
        .annotate(query_similarity=TrigramWordSimilarity(query, "u_query"))
        .annotate(
            max_similarity=Greatest(
                "artist_similarity", "title_similarity", "query_similarity"
            )
        )
        .values(*u_values_list, "u_query", "max_similarity")
    )

    query_result = similar_songs.union(similar_queries)

    query_result = query_result.order_by("-max_similarity", "u_artist", "u_title")[
        : storage.get("number_of_suggestions")
    ]

    # The selected values need to contain the field by which the set is ordered.
    # Some entries are duplicated if they have different (but high) similarities.
    # Distinct is not allowed after a union.
    # Thus, duplicates need to be eliminated in python separately.
    song_results = list({song["u_id"]: song for song in query_result}.values())

    # This is the same query without using union.
    # It does not use indexes and is thus a lot slower.
    # similar_songs = (
    #   ArchivedSong.objects.select_related()
    #   .filter(
    #       Q(artist__trigram_word_similar=query)
    #       | Q(title__trigram_word_similar=query)
    #       | Q(queries__query__trigram_word_similar=query)
    #   )
    #   .annotate(artist_similarity=TrigramWordSimilarity(query, "artist"))
    #   .annotate(title_similarity=TrigramWordSimilarity(query, "title"))
    #   .annotate(
    #       query_similarity=Coalesce(
    #           TrigramWordSimilarity(query, "queries__query"), 0
    #       )
    #   )
    #   .annotate(
    #       max_similarity=Greatest(
    #           "artist_similarity", "title_similarity", "query_similarity"
    #       )
    #   )
    #   .values(*values_list, "max_similarity")
    #   .order_by("-max_similarity", "artist", "title")[:20]
    # )

    return song_results


def _offline_song_suggestions(query: str) -> List[SuggestionResult]:
    results: List[SuggestionResult] = []
    terms = query.split()
    song_results: Iterable[Mapping[str, Any]]
    if settings.DEBUG:
        # Testing the whole table whether it contains any term is quite costly.
        # Used for sqlite3 which does not have a similarity function.
        matching_songs = ArchivedSong.objects.prefetch_related("queries")
        for term in terms:
            matching_songs = matching_songs.filter(
                Q(title__icontains=term)
                | Q(artist__icontains=term)
                | Q(queries__query__icontains=term)
            )

        song_results = (
            matching_songs.order_by("-counter")
            # annotate with same values as in the postgres case to have a consistent interface
            .annotate(
                u_id=F("id"),
                u_url=F("url"),
                u_artist=F("artist"),
                u_title=F("title"),
                u_duration=F("duration"),
                u_counter=F("counter"),
                u_cached=F("cached"),
            )
            .values(*u_values_list)
            .distinct()[: storage.get("number_of_suggestions")]
        )

        # Perhaps this could be combined with the similarity search
        # to improve usability with the right weighting.
        # matching_songs = matching_songs.annotate(
        #    artist_similarity=Coalesce(TrigramWordSimilarity(query, "artist"), 0),
        #    title_similarity=Coalesce(TrigramWordSimilarity(query, "title"), 0),
        #    query_similarity=Coalesce(TrigramWordSimilarity(query, "queries__query"), 0),
        #    max_similarity=Greatest(
        #        "artist_similarity", "title_similarity", "query_similarity"
        #    ),
        # )

        # To combine, use union instead of | (or) in order to access the annotated values
        # similar_songs = union(matching_songs)
    else:
        song_results = _postgres_song_results(query)

    has_internet = redis.get("has_internet")
    for song in song_results:
        if song_utils.is_forbidden(song["u_artist"]) or song_utils.is_forbidden(
            song["u_title"]
        ):
            continue

        platform = song_utils.determine_url_type(song["u_url"])
        # don't suggest online songs when we don't have internet
        if not has_internet and not song["u_cached"]:
            continue
        if platform == "local":
            # don't suggest local songs if they are not cached (=not at expected location)
            if not song["u_cached"]:
                continue
        else:
            # don't suggest songs if the respective platform is disabled
            assert platform in ["youtube", "spotify", "soundcloud", "jamendo"]
            if not storage.get(cast(PlatformEnabled, f"{platform}_enabled")):
                continue
        result_dict: SuggestionResult = {
            "key": song["u_id"],
            "value": song_utils.displayname(song["u_artist"], song["u_title"]),
            "counter": song["u_counter"],
            "type": platform,
            "durationFormatted": song_utils.format_seconds(song["u_duration"]),
        }
        results.append(result_dict)
    # mark suggestions whose displayname is identical
    seen_values: Dict[str, int] = {}
    for index, result in enumerate(results):
        if result["value"] in seen_values:
            result["confusable"] = True
            results[seen_values[result["value"]]]["confusable"] = True
        seen_values[result["value"]] = index
    return results


def offline_suggestions(request: WSGIRequest) -> JsonResponse:
    """Returns offline suggestions for a given query."""
    query = request.GET["term"]
    suggest_playlist = request.GET["playlist"] == "true"

    if storage.get("new_music_only"):
        return JsonResponse([], safe=False)

    if suggest_playlist:
        results = _offline_playlist_suggestions(query)
    else:
        results = _offline_song_suggestions(query)

    return JsonResponse(results, safe=False)



"""This module interfaces with mopidy."""

import logging
import urllib.parse
from contextlib import contextmanager
from typing import Iterator, Optional
from threading import Event

import requests
from django.conf import settings as conf
from mopidyapi.client import MopidyAPI
from mopidyapi.exceptions import MopidyError


from core import redis
from core.musiq.playback import PlaybackError
from core.settings import storage
from core.musiq import player

# this lock is released when restarting mopidy (which happens in another Thread)
mopidy_lock = redis.connection.lock("mopidy_lock", thread_local=False)

# this creates two PLAYER objects, once in the playback worker
# and once in the request handler
# during testing, both are the same process and thus only one instance is created
PLAYER = MopidyAPI(host=conf.MOPIDY_HOST, port=conf.MOPIDY_PORT)


@contextmanager
def mopidy_command(important: bool = False) -> Iterator[bool]:
    """A context that should be used around every mopidy command used.
    Makes sure that commands occur sequentially, as mopidy can not handle parallel inputs.
    Use it like this:
    with mopidy_command() as allowed:
        if allowed:
            # mopidy command
    :param important: If True, wait until the lock is released.
    If not, return 'False' after a timeout."""
    timeout: Optional[int] = 3
    if important:
        timeout = None
    if mopidy_lock.acquire(blocking_timeout=timeout):
        yield True
        mopidy_lock.release()
    else:
        logging.warning("mopidy command could not be executed")
        player.set_playback_error(True)
        yield False


class MopidyPlayer(player.Player):
    """Class containing methods to interface with mopidy."""

    def __init__(self):
        self.playback_started = Event()

        with mopidy_command(important=True):
            PLAYER.playback.stop()
            PLAYER.tracklist.clear()
            # make songs disappear from tracklist after being played
            PLAYER.tracklist.set_consume(True)

        @PLAYER.on_event("track_playback_started")
        def _on_playback_started(_event) -> None:
            self.playback_started.set()

    def start_song(self, song, catch_up: float):
        with mopidy_command(important=True):
            # after a restart consume may be set to False again, so make sure it is on
            PLAYER.tracklist.clear()
            PLAYER.tracklist.set_consume(True)
            PLAYER.tracklist.add(uris=[song.internal_url])
            # temporarily mute mopidy in case we need to seek but mopidy does not react directly
            # this allows us to seek first and then unmute, preventing audible skips
            volume = PLAYER.mixer.get_volume()
            if catch_up is not None and catch_up >= 0:
                PLAYER.mixer.set_volume(0)
            # mopidy can only seek when the song is playing
            # also we do not continue without the playing state properly set.
            # otherwise waiting might exit before the song started
            PLAYER.playback.play()
            if not self.playback_started.wait(timeout=1):
                # mopidy did not acknowledge that it started the song
                # to make sure it is not in an error state,
                # restart the loop and retry to start the song
                # also prevents "queue-eating" bug,
                # where mopidy in a failed state would refuse to play any song,
                # but raveberry keeps on sending songs from the queue
                logging.warning("playback_started event did not trigger")
                player.set_playback_error(True)
                PLAYER.mixer.set_volume(volume)
                raise PlaybackError("playback_started event did not trigger")
            player.set_playback_error(False)
            if catch_up is not None and catch_up >= 0:
                # instead of seeking to the specified time
                # mopidy sometimes just starts the song from the beginning
                # checking the time position before and after seeking
                # prevents this from happening for some reason
                PLAYER.playback.get_time_position()
                PLAYER.playback.seek(catch_up)
                PLAYER.playback.get_time_position()
                if redis.get("paused"):
                    PLAYER.playback.pause()
                PLAYER.mixer.set_volume(volume)

    def should_stop_waiting(self, previous_error: bool) -> bool:
        error = False
        with mopidy_command() as allowed:
            if allowed:
                try:
                    # if there was an error previously, then a stopped state means
                    # that mopidy is back up. stop waiting and continue the main loop
                    if PLAYER.playback.get_state() == "stopped" and previous_error:
                        return True
                    # previously we also used this to check for the song end,
                    # but this is now done via the start point and duration of a song
                except (requests.exceptions.ConnectionError, MopidyError):
                    # error during state get, skip until reconnected
                    error = True
        if error:
            # raise outside of the with block to free resources
            raise PlaybackError
        return False

    def play_alarm(self, interrupt: bool, alarm_path: str) -> None:
        self.playback_started.clear()
        with mopidy_command(important=True):
            # interrupt the current song if its playing
            if interrupt:
                PLAYER.tracklist.clear()

            PLAYER.tracklist.add(uris=["file://" + urllib.parse.quote(alarm_path)])
            PLAYER.playback.play()
        self.playback_started.wait(timeout=1)

    def play_backup_stream(self):
        PLAYER.tracklist.add(uris=[storage.get("backup_stream")])
        PLAYER.playback.play()


def restart() -> None:
    with mopidy_command() as allowed:
        if allowed:
            PLAYER.playback.seek(0)


def seek_backward(seek_distance: float) -> None:
    with mopidy_command() as allowed:
        if allowed:
            current_position = PLAYER.playback.get_time_position()
            PLAYER.playback.seek(current_position - seek_distance * 1000)


def play() -> None:
    with mopidy_command() as allowed:
        if allowed:
            PLAYER.playback.play()


def pause() -> None:
    with mopidy_command() as allowed:
        if allowed:
            PLAYER.playback.pause()


def seek_forward(seek_distance: float) -> None:
    with mopidy_command() as allowed:
        if allowed:
            current_position = PLAYER.playback.get_time_position()
            PLAYER.playback.seek(current_position + seek_distance * 1000)


def skip() -> None:
    with mopidy_command() as allowed:
        if allowed:
            PLAYER.playback.next()


def set_volume(volume) -> None:
    # change mopidy's volume
    with mopidy_command() as allowed:
        if allowed:
            PLAYER.mixer.set_volume(round(volume * 100))



"""This module contains all public api endpoints."""
import re

from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse, HttpResponseBadRequest
from django.views.decorators.csrf import csrf_exempt

from django.conf import settings as conf
from core.musiq import musiq
from core.musiq.music_provider import ProviderError


@csrf_exempt
def post_song(request: WSGIRequest) -> HttpResponse:
    """This endpoint is part of the API and exempt from CSRF checks.
    Shareberry uses this endpoint."""
    # only get ip on user requests
    query = request.POST.get("query")
    if not query:
        return HttpResponseBadRequest("No query to share.")

    match = re.search(r"(?P<url>https?://[^\s]+)", query)
    if match:
        query = match.group("url")

    try:
        providers = musiq.get_providers(query)
    except ProviderError as error:
        return HttpResponseBadRequest(str(error))
    provider = musiq.try_providers(request.session.session_key, providers)
    if provider.error:
        return HttpResponseBadRequest(provider.error)
    return HttpResponse(provider.ok_message)


def version(request: WSGIRequest) -> HttpResponse:
    """Return the version of the running instance."""

    return HttpResponse(f"Raveberry version {conf.VERSION}")



"""Controls the Raspberry Pi's internal LEDs."""
import os
import subprocess


def _control_led(led: str, action: str) -> None:
    control_led = "/usr/local/sbin/raveberry/control_led"
    if os.path.isfile(control_led):
        subprocess.call(["sudo", control_led, led, action], stderr=subprocess.DEVNULL)


def enable_act_led() -> None:
    """Enables the green activity led on a Raspberry Pi."""
    _control_led("act", "enable")


def disable_act_led() -> None:
    """Disables the green activity led on a Raspberry Pi."""
    _control_led("act", "disable")


def enable_pwr_led() -> None:
    """Enables the red power led on a Raspberry Pi."""
    _control_led("pwr", "enable")


def disable_pwr_led() -> None:
    """Disables the red power led on a Raspberry Pi."""
    _control_led("pwr", "disable")



"""This module handles all basic settings."""

from __future__ import annotations

import subprocess

from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse, HttpResponseBadRequest

from core import redis, user_manager
from core.musiq import playback
from core.settings import storage
from core.settings.settings import control
from core.util import strtobool, extract_value


def start() -> None:
    """Initializes this module. Checks whether internet is accessible."""
    _check_internet()


def _check_internet() -> None:
    host = storage.get("connectivity_host")
    if not host:
        redis.put("has_internet", False)
        return
    response = subprocess.call(
        ["ping", "-c", "1", "-W", "3", host], stdout=subprocess.DEVNULL
    )
    if response == 0:
        redis.put("has_internet", True)
    else:
        redis.put("has_internet", False)


@control
def set_interactivity(request: WSGIRequest) -> HttpResponse:
    """Enables or disables voting based on the given value."""
    value, response = extract_value(request.POST)
    if value not in [
        getattr(storage.Interactivity, attr)
        for attr in dir(storage.Interactivity)
        if not attr.startswith("__")
    ]:
        return HttpResponseBadRequest("Invalid value")
    storage.put("interactivity", value)
    return response


@control
def set_color_indication(request: WSGIRequest) -> HttpResponse:
    """Enables or disables voting based on the given value."""
    value, response = extract_value(request.POST)
    if value not in [
        getattr(storage.Privileges, attr)
        for attr in dir(storage.Privileges)
        if not attr.startswith("__")
    ]:
        return HttpResponseBadRequest("Invalid value")
    storage.put("color_indication", value)
    return response


@control
def set_ip_checking(request: WSGIRequest) -> HttpResponse:
    """Enables or disables ip checking based on the given value."""
    value, response = extract_value(request.POST)
    storage.put("ip_checking", strtobool(value))
    return response


@control
def set_downvotes_to_kick(request: WSGIRequest) -> HttpResponse:
    """Sets the number of downvotes that are needed to remove a song from the queue."""
    value, response = extract_value(request.POST)
    storage.put("downvotes_to_kick", int(value))
    return response


@control
def set_logging_enabled(request: WSGIRequest) -> HttpResponse:
    """Enables or disables logging of requests and play logs based on the given value."""
    value, response = extract_value(request.POST)
    storage.put("logging_enabled", strtobool(value))
    return response


@control
def set_hashtags_active(request: WSGIRequest) -> HttpResponse:
    """Enables or disables logging of requests and play logs based on the given value."""
    value, response = extract_value(request.POST)
    storage.put("hashtags_active", strtobool(value))
    return response


@control
def set_privileged_stream(request: WSGIRequest) -> HttpResponse:
    """Enables or disables logging of requests and play logs based on the given value."""
    value, response = extract_value(request.POST)
    storage.put("privileged_stream", strtobool(value))
    return response


@control
def set_online_suggestions(request: WSGIRequest) -> HttpResponse:
    """Enables or disables online suggestions based on the given value."""
    value, response = extract_value(request.POST)
    storage.put("online_suggestions", strtobool(value))
    return response


@control
def set_number_of_suggestions(request: WSGIRequest) -> HttpResponse:
    """Set the number of archived suggestions based on the given value."""
    value, response = extract_value(request.POST)
    storage.put("number_of_suggestions", int(value))
    return response


@control
def set_connectivity_host(request: WSGIRequest) -> HttpResponse:
    """Sets the host that is pinged to check the internet connection."""
    value, response = extract_value(request.POST)
    storage.put("connectivity_host", value)
    return response


@control
def check_internet(_request: WSGIRequest) -> None:
    """Checks whether an internet connection exists and updates the internal state."""
    _check_internet()


@control
def update_user_count(_request: WSGIRequest) -> None:
    """Force an update on the active user count."""
    user_manager.update_user_count()


@control
def set_new_music_only(request: WSGIRequest) -> HttpResponse:
    """Enables or disables the new music only mode based on the given value."""
    value, response = extract_value(request.POST)
    storage.put("new_music_only", strtobool(value))
    return response


@control
def set_enqueue_first(request: WSGIRequest) -> HttpResponse:
    """Enables or disables the new music only mode based on the given value."""
    value, response = extract_value(request.POST)
    storage.put("enqueue_first", strtobool(value))
    return response


@control
def set_song_cooldown(request: WSGIRequest) -> HttpResponse:
    """Enables or disables the new music only mode based on the given value."""
    value, response = extract_value(request.POST)
    storage.put("song_cooldown", float(value))
    return response


@control
def set_max_download_size(request: WSGIRequest) -> HttpResponse:
    """Sets the maximum amount of MB that are allowed for a song that needs to be downloaded."""
    value, response = extract_value(request.POST)
    storage.put("max_download_size", float(value))
    return response


@control
def set_max_playlist_items(request: WSGIRequest) -> HttpResponse:
    """Sets the maximum number of songs that are downloaded from a playlist."""
    value, response = extract_value(request.POST)
    storage.put("max_playlist_items", int(value))
    return response


@control
def set_max_queue_length(request: WSGIRequest) -> HttpResponse:
    """Sets the maximum number of songs that are downloaded from a playlist."""
    value, response = extract_value(request.POST)
    storage.put("max_queue_length", int(value))
    return response


@control
def set_additional_keywords(request: WSGIRequest) -> HttpResponse:
    """Sets the keywords to filter out of results."""
    value, response = extract_value(request.POST)
    storage.put("additional_keywords", value)
    return response


@control
def set_forbidden_keywords(request: WSGIRequest) -> HttpResponse:
    """Sets the keywords to filter out of results."""
    value, response = extract_value(request.POST)
    storage.put("forbidden_keywords", value)
    return response


@control
def set_people_to_party(request: WSGIRequest) -> HttpResponse:
    """Sets the amount of active clients needed to enable partymode."""
    value, response = extract_value(request.POST)
    storage.put("people_to_party", int(value))
    return response


@control
def set_alarm_probability(request: WSGIRequest) -> HttpResponse:
    """Sets the probability with which an alarm is triggered after each song."""
    value, response = extract_value(request.POST)
    storage.put("alarm_probability", float(value))
    return response


@control
def set_buzzer_cooldown(request: WSGIRequest) -> HttpResponse:
    """Sets the minimum time that needs to pass between buzzer presses."""
    value, response = extract_value(request.POST)
    storage.put("buzzer_cooldown", float(value))
    return response


@control
def set_buzzer_success_probability(request: WSGIRequest) -> HttpResponse:
    """Sets the probability for the buzzer to play a success sound."""
    value, response = extract_value(request.POST)
    storage.put("buzzer_success_probability", float(value))
    return response


@control
def trigger_alarm(_request: WSGIRequest) -> None:
    """Manually triggers an alarm."""
    if playback.trigger_alarm():
        # because a state update is sent after every control (including this one)
        # a state update with alarm not being set would be sent
        # prevent this by manually setting this redis variable prematurely
        redis.put("alarm_playing", True)



# Generated by Django 2.2.24 on 2021-11-28 18:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("core", "0015_install_trgm_extension")]

    operations = [
        migrations.AddField(
            model_name="archivedsong",
            name="cached",
            field=models.BooleanField(default=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="archivedsong",
            name="duration",
            # with a default of -1,
            # existing uncached songs (e.g. spotify) will have an unknown duration.
            # To fix this, run a one-off script that fetches the duration for all old uncached songs
            # New songs are not affected, they will have the correct duration.
            field=models.FloatField(default=-1),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="currentsong", name="duration", field=models.FloatField()
        ),
        migrations.AlterField(
            model_name="queuedsong", name="duration", field=models.FloatField()
        ),
    ]



"""This module contains all programs that use leds."""
import colorsys
import math
from typing import List, Tuple, cast, TYPE_CHECKING

from core.lights.programs import LedProgram

if TYPE_CHECKING:
    from core.lights.worker import DeviceManager


def stretched_hues(led_count: int, offset: float = 0):
    """Stretches red and blue, compresses green and pink."""
    # Uses the logistic curve to make colors more prominent and compress the others
    #
    #  ^ out hue
    # 1-        xx
    #  |       x
    #  |     xx
    #  |    x
    #  |   x
    #  |   x
    #  |  x
    #  |xx
    #  -|--|--|--|> in hue
    #   R  G  B  R
    #
    # Two logistic curves are combined to compress two colors (green and pink).
    # Green is compressed because the board is green and visible anyway.
    # Pink just does not look that good.

    max1 = 2 / 3
    max2 = 1 / 3

    def logistic(fraction: float) -> float:
        # First curve, compresses green (hue = ⅓)
        def logistic1(val: float) -> float:
            return max1 / (1 + math.e ** (-16 * (val - 1 / 3)))

        # First curve, compresses pink (hue = ⅚)
        def logistic2(val: float) -> float:
            return max2 / (1 + math.e ** (-16 * (val - 5 / 6)))

        if fraction < 2 / 3:
            # Vertically stretch and move the curve so it starts at y=0 and ends at y=M
            yoffset = logistic1(0)
            scale = max1 / (max1 - 2 * yoffset)
            return scale * (logistic1(fraction) - yoffset)

        yoffset = logistic2(2 / 3)
        scale = max2 / (max2 - 2 * yoffset)
        return scale * (logistic2(fraction) - yoffset) + max1

    return [logistic((offset + led / led_count) % 1) % 1 for led in range(0, led_count)]


def stretched_hues_spectrum(led_count: int):
    """Stretches red and blue, compressing green, but removes pink.
    Adds a short red section, because red is chronically underrepresented.
    Doesn't take an offset, because the ends do not match up,
    leading to jumps in hue. Only used for the spectrum."""
    #  ^ out hue
    # 1-
    #  |
    # ⅔-       xx
    #  |      x
    #  |     x
    #  |     x
    #  |    x
    #  |xxxx
    #  -|--|--|--|> in hue
    #   R  G  B  R
    max_value = 2 / 3

    def scaled_logistic(fraction: float) -> float:
        def logistic(val: float) -> float:
            return max_value / (1 + math.e ** (-12 * (val - 9 / 16)))

        if fraction < 1 / 8:
            return 0
        yoffset = logistic(1 / 8)
        scale = max_value / (max_value - 2 * yoffset)
        return scale * logistic(fraction) - yoffset

    return [scaled_logistic(led / led_count) % 1 for led in range(0, led_count)]


class Fixed(LedProgram):
    """Show one fixed color only. The color is controlled in the "DeviceManager" class."""

    def __init__(self, manager: "DeviceManager") -> None:
        super().__init__(manager, "Fixed")

    def compute(self) -> None:
        # show a red color if the alarm is active
        alarm_factor = self.manager.utilities.alarm.factor
        if alarm_factor != -1.0:
            self.manager.settings["fixed_color"] = (alarm_factor, 0, 0)

    def ring_colors(self) -> List[Tuple[float, float, float]]:
        return [
            self.manager.settings["fixed_color"]
            for _ in range(self.manager.devices.ring.LED_COUNT)
        ]

    def wled_colors(self) -> List[Tuple[float, float, float]]:
        return [
            self.manager.settings["fixed_color"]
            for _ in range(self.manager.devices.wled.led_count)
        ]

    def strip_color(self) -> Tuple[float, float, float]:
        return self.manager.settings["fixed_color"]


class Rainbow(LedProgram):
    """Continuously cycles through all colors. Affected by the speed setting."""

    def __init__(self, manager: "DeviceManager") -> None:
        super().__init__(manager, "Rainbow")
        self.program_duration = 1
        self.time_passed = 0.0
        self.current_fraction = 0.0

    def start(self) -> None:
        self.time_passed = 0.0

    def compute(self) -> None:
        self.time_passed += (
            1 / self.manager.settings["ups"] * self.manager.settings["program_speed"]
        )
        self.time_passed %= self.program_duration
        self.current_fraction = self.time_passed / self.program_duration

    def _colors(self, led_count) -> List[Tuple[float, float, float]]:
        return [
            colorsys.hsv_to_rgb(hue, 1, 1)
            for hue in stretched_hues(led_count, self.current_fraction)
        ]

    def ring_colors(self) -> List[Tuple[float, float, float]]:
        return self._colors(self.manager.devices.ring.LED_COUNT)

    def wled_colors(self) -> List[Tuple[float, float, float]]:
        return self._colors(self.manager.devices.wled.led_count)

    def strip_color(self) -> Tuple[float, float, float]:
        return colorsys.hsv_to_rgb(self.current_fraction, 1, 1)


class Adaptive(LedProgram):
    """Dynamically reacts to the currently played music.
    Low frequencies are represented by red, high ones by blue."""

    def __init__(self, manager: "DeviceManager") -> None:
        super().__init__(manager, "Rave")
        self.cava = self.manager.utilities.cava

        # RING
        # The spectrum needs to have a color for low frequencies (red)
        # and a color for high frequencies (blue)
        # In order to show a clean separation between the spectrum ends,
        # the color between the two (pink) is removed from the pool of possible colors.
        ring_hues = stretched_hues_spectrum(self.manager.devices.ring.LED_COUNT)
        self.ring_base_colors = [colorsys.hsv_to_rgb(hue, 1, 1) for hue in ring_hues]

        # WLED
        # identical to ring, but with a different number of leds
        wled_hues = stretched_hues_spectrum(self.manager.devices.wled.led_count)
        self.wled_base_colors = [colorsys.hsv_to_rgb(hue, 1, 1) for hue in wled_hues]

        # STRIP
        # distribute frequencies over the three leds. Don't use hard cuts, but smooth functions
        # the functions add up to one at every point and each functions integral is a third
        self.strip_granularity = 16
        self.red_coeffs = [
            -1
            / (
                1
                + math.e ** (-6 * math.e * (led / (self.strip_granularity - 1) - 1 / 3))
            )
            + 1
            for led in range(0, self.strip_granularity)
        ]
        self.blue_coeffs = [
            1
            / (
                1
                + math.e ** (-6 * math.e * (led / (self.strip_granularity - 1) - 2 / 3))
            )
            for led in range(0, self.strip_granularity)
        ]
        self.green_coeffs = [
            1 - self.red_coeffs[led] - self.blue_coeffs[led]
            for led in range(0, self.strip_granularity)
        ]

    def start(self) -> None:
        self.cava.use()

    def compute(self) -> None:
        pass

    def _aggregate_frame(self, led_count) -> List[float]:
        # aggregate the length of cavas frame into a list the length of the number of leds we have.
        # This reduces computation time.
        values_per_led = len(self.cava.current_frame) // led_count
        left = len(self.cava.current_frame) - values_per_led * led_count

        start = 0

        aggregated = []
        for _ in range(led_count):
            end = start + values_per_led
            bin_size = values_per_led
            if left > 0:
                end += 1
                left -= 1
                bin_size += 1

            aggregated.append(sum(self.cava.current_frame[start:end]) / bin_size)
            start += values_per_led
        return aggregated

    def ring_colors(self) -> List[Tuple[float, float, float]]:
        aggregated = self._aggregate_frame(self.manager.devices.ring.LED_COUNT)
        colors = [
            tuple(factor * val for val in color)
            for factor, color in zip(aggregated, self.ring_base_colors)
        ]
        # https://github.com/python/mypy/issues/5068
        return cast(List[Tuple[float, float, float]], colors)

    def wled_colors(self) -> List[Tuple[float, float, float]]:
        aggregated = self._aggregate_frame(self.manager.devices.wled.led_count)
        colors = [
            tuple(factor * val for val in color)
            for factor, color in zip(aggregated, self.wled_base_colors)
        ]
        # https://github.com/python/mypy/issues/5068
        return cast(List[Tuple[float, float, float]], colors)

    def strip_color(self) -> Tuple[float, float, float]:
        aggregated = self._aggregate_frame(self.strip_granularity)
        red = (
            sum(coeff * val for coeff, val in zip(self.red_coeffs, aggregated))
            * 3
            / self.strip_granularity
        )
        green = (
            sum(coeff * val for coeff, val in zip(self.green_coeffs, aggregated))
            * 3
            / self.strip_granularity
        )
        blue = (
            sum(coeff * val for coeff, val in zip(self.blue_coeffs, aggregated))
            * 3
            / self.strip_granularity
        )
        red = min(1.0, red)
        green = min(1.0, green)
        blue = min(1.0, blue)
        return red, green, blue

    def stop(self) -> None:
        self.cava.release()



from typing import Dict, List, Literal, Tuple, overload

from redis import Redis

DeviceInitialized = Literal[
    "ring_initialized", "wled_initialized", "strip_initialized", "screen_initialized"
]

connection: Redis

def start() -> None: ...

class Event:
    def __init__(self, name: str) -> None: ...
    def wait(self) -> None: ...
    def set(self) -> None: ...
    def clear(self) -> None: ...

@overload
def get(
    key: Literal[
        "playing",
        "paused",
        "playback_error",
        "stop_playback_loop",
        "alarm_playing",
        "alarm_requested",
        "backup_playing",
        "lights_active",
        "ring_initialized",
        "wled_initialized",
        "strip_initialized",
        "screen_initialized",
        "has_internet",
        "mopidy_available",
        "youtube_available",
        "spotify_available",
        "soundcloud_available",
        "jamendo_available",
        "bluetoothctl_active",
    ]
) -> bool: ...
@overload
def get(key: Literal["active_requests"]) -> int: ...
@overload
def get(
    key: Literal["alarm_duration", "last_buzzer", "current_fps", "last_user_count_update"]
) -> float: ...
@overload
def get(key: Literal["active_player", "library_scan_progress"]) -> str: ...
@overload
def get(key: Literal["led_programs", "screen_programs"]) -> List[str]: ...
@overload
def get(key: Literal["resolutions"]) -> List[Tuple[int, int]]: ...
@overload
def get(key: Literal["current_resolution"]) -> Tuple[int, int]: ...
@overload
def get(key: Literal["bluetooth_devices"]) -> List[Dict[str, str]]: ...
@overload
def get(key: Literal["last_requests"]) -> Dict[str, float]: ...
@overload
def put(
    key: Literal[
        "playing",
        "paused",
        "playback_error",
        "stop_playback_loop",
        "alarm_playing",
        "alarm_requested",
        "backup_playing",
        "lights_active",
        "ring_initialized",
        "wled_initialized",
        "strip_initialized",
        "screen_initialized",
        "has_internet",
        "mopidy_available,
        "youtube_available",
        "spotify_available",
        "soundcloud_available",
        "jamendo_available",
        "bluetoothctl_active",
    ],
    value: bool,
) -> None: ...
@overload
def put(key: Literal["active_requests"], value: int) -> None: ...
@overload
def put(
    key: Literal["alarm_duration", "last_buzzer", "current_fps", "last_user_count_update"],
    value: float,
) -> None: ...
@overload
def put(key: Literal["active_player", "library_scan_progress"], value: str) -> None: ...
@overload
def put(key: Literal["led_programs", "screen_programs"], value: List[str]) -> None: ...
@overload
def put(key: Literal["resolutions"], value: List[Tuple[int, int]]) -> None: ...
@overload
def put(key: Literal["current_resolution"], value: Tuple[int, int]) -> None: ...
@overload
def put(key: Literal["bluetooth_devices"], value: List[Dict[str, str]]) -> None: ...
@overload
def put(key: Literal["last_requests"], value: Dict[str, float]) -> None: ...



"""This module contains all Spotify related code."""

from __future__ import annotations

import json
from typing import List, Optional, Tuple
from urllib.parse import urlparse

import spotipy
from django.http.response import HttpResponse
from spotipy import SpotifyOAuth, SpotifyStateError

from core.musiq import song_utils
from core.musiq.playlist_provider import PlaylistProvider
from core.musiq.song_provider import SongProvider
from core.musiq.spotify_web import OAuthClient
from core.settings import storage
from core import redis


class staticproperty(property):
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()


def _get_auth_response_storage(self, open_browser=False):
    response = storage.get("spotify_authorized_url")
    state, code = SpotifyOAuth.parse_auth_response_url(response)
    if self.state is not None and self.state != state:
        raise SpotifyStateError(self.state, state)
    return code


class DatabaseCacheHandler(spotipy.CacheHandler):
    """A cache handler for spotipy OAuth that stores the token info in the database."""

    def get_cached_token(self):
        token_info = storage.get("spotipy_token_info")
        if not token_info:
            return None
        return json.loads(token_info)

    def save_token_to_cache(self, token_info):
        storage.put("spotipy_token_info", json.dumps(token_info))


class Spotify:
    """This class contains code for both the song and playlist provider"""

    _device_api = None  # type: ignore[assignment]
    _mopidy_api = None  # type: ignore[assignment]

    @staticmethod
    def device_api():
        client_id = storage.get(key="spotify_device_client_id")
        client_secret = storage.get(key="spotify_device_client_secret")
        redirect_uri = storage.get(key="spotify_redirect_uri")
        scope = "user-read-playback-state user-modify-playback-state playlist-read-private playlist-read-collaborative user-library-read"

        auth_manager = SpotifyOAuth(
            client_id=client_id,
            client_secret=client_secret,
            redirect_uri=redirect_uri,
            scope=scope,
            cache_handler=DatabaseCacheHandler(),
        )
        auth_manager._get_auth_response_interactive = (
            _get_auth_response_storage.__get__(auth_manager)
        )

        return spotipy.Spotify(auth_manager=auth_manager)

    @staticmethod
    def mopidy_api():
        client_id = storage.get(key="spotify_mopidy_client_id")
        client_secret = storage.get(key="spotify_mopidy_client_secret")
        return OAuthClient(
            base_url="https://api.spotify.com/v1",
            refresh_url="https://auth.mopidy.com/spotify/token",
            client_id=client_id,
            client_secret=client_secret,
        )

    @staticmethod
    def create_device_api():
        Spotify._device_api = Spotify.device_api()

    @staticmethod
    def create_mopidy_api():
        Spotify._mopidy_api = Spotify.mopidy_api()

    @staticproperty
    def api(cls):
        """Returns the spotify client if it was already created.
        If not, it is created using the spotify credentials from the database."""
        if redis.get("active_player") == "spotify":
            if cls._device_api is None:
                cls.create_device_api()
            return cls._device_api
        elif redis.get("active_player") == "mopidy":
            if cls._mopidy_api is None:
                cls.create_mopidy_api()
            return cls._mopidy_api

    def get_search_suggestions(
        self, query: str, playlist: bool
    ) -> List[Tuple[str, str]]:
        """Returns a list of suggested items for the given query.
        Returns playlists if :param playlist: is True, songs otherwise."""
        result = self.api.search(
            query, limit=20, type="playlist" if playlist else "track"
        )

        if playlist:
            items = result["playlists"]["items"]
        else:
            items = result["tracks"]["items"]

        suggestions = []
        for item in items:
            external_url = item["external_urls"]["spotify"]
            title = item["name"]
            if playlist:
                displayname = title
            else:
                artist = item["artists"][0]["name"]
                # apply filter from the settings
                if song_utils.is_forbidden(artist) or song_utils.is_forbidden(title):
                    continue
                displayname = song_utils.displayname(artist, title)
            suggestions.append((displayname, external_url))

        # remove duplicates and filter by keywords
        chosen_displaynames = set()
        unique_suggestions = []
        for suggestion in suggestions:
            if suggestion[0] in chosen_displaynames:
                continue
            unique_suggestions.append(suggestion)
            chosen_displaynames.add(suggestion[0])
        return unique_suggestions


class SpotifySongProvider(SongProvider, Spotify):
    """This class handles songs from Spotify."""

    @staticmethod
    def get_id_from_external_url(url: str) -> str:
        return urlparse(url).path.split("/")[-1]

    @staticmethod
    def get_id_from_internal_url(url: str) -> str:
        """Returns the internal id based on the given url."""
        return url.split(":")[-1]

    def __init__(self, query: Optional[str], key: Optional[int]) -> None:
        self.type = "spotify"
        super().__init__(query, key)

    def check_available(self) -> bool:
        if not self.gather_metadata():
            return False
        # the default bitrate of mopidy-spotify is 160kbps
        # estimate the size of a song by multiplying with its duration
        size = self.metadata["duration"] * 160 / 8 * 1000
        return self.check_not_too_large(size)

    def gather_metadata(self) -> bool:
        """Fetches metadata for this song's uri from Spotify."""
        if not self.id:
            results = self.api.search(self.query, type="track")

            result = self.first_unfiltered_item(
                results["tracks"]["items"],
                lambda item: (item["artists"][0]["name"], item["name"]),
            )
            if not result:
                return False
            self.id = result["id"]
        else:
            result = self.api.track(self.id)
        assert result
        try:
            self.metadata["artist"] = result["artists"][0]["name"]
            self.metadata["title"] = result["name"]
            self.metadata["duration"] = result["duration_ms"] / 1000
            self.metadata["internal_url"] = result["uri"]
            self.metadata["external_url"] = result["external_urls"]["spotify"]
            self.metadata["stream_url"] = None
            self.metadata["cached"] = False
        except KeyError:
            self.error = "No song found"
            return False
        return True

    def get_internal_url(self) -> str:
        if not self.id:
            raise ValueError()
        return "spotify:track:" + self.id

    def get_external_url(self) -> str:
        if not self.id:
            raise ValueError()
        return "https://open.spotify.com/track/" + self.id

    def get_suggestion(self) -> str:
        result = self.api.recommendations(limit=1, seed_tracks=[self.id])

        try:
            external_url = result["tracks"][0]["external_urls"]["spotify"]
        except (IndexError, KeyError) as error:
            self.error = "no recommendation found"
            raise ValueError("No suggested track") from error

        return external_url

    def request_radio(self, session_key: str) -> HttpResponse:
        result = self.api.recommendations(
            limit=storage.get("max_playlist_items"), seed_tracks=[self.id]
        )

        for track in result["tracks"]:
            external_url = track["external_urls"]["spotify"]
            provider = SpotifySongProvider(external_url, None)
            provider.request("", archive=False, manually_requested=False)

        return HttpResponse("queueing radio")


class SpotifyPlaylistProvider(PlaylistProvider, Spotify):
    """This class handles Spotify Playlists."""

    @staticmethod
    def get_id_from_external_url(url: str) -> Optional[str]:
        if not (
            url.startswith("https://open.spotify.com/playlist/")
            or url.startswith("https://open.spotify.com/artist/")
            or url.startswith("https://open.spotify.com/album/")
        ):
            return None
        return urlparse(url).path.split("/")[-1]

    def __init__(self, query: Optional[str], key: Optional[int]) -> None:
        self.type = "spotify"
        # can be one of playlists, artists or albums
        # defaults to playlist, only changes if an artist or album is found during search_id
        # this type is not reflected in list_id. Thus, ArchivedPlaylist entries do not know
        # what kind of collection of songs they are.
        # This is considered acceptable, generating external urls from playlist is never required
        # and finding cached lists still works as extracted ids still match
        self._spotify_type = "playlist"
        if query:
            if query.startswith("https://open.spotify.com/playlist/"):
                self._spotify_type = "playlist"
            elif query.startswith("https://open.spotify.com/artist/"):
                self._spotify_type = "artist"
            elif query.startswith("https://open.spotify.com/album/"):
                self._spotify_type = "album"
            super().__init__(query, key)

    def search_id(self) -> Optional[str]:
        result = self.api.search(self.query, limit=1, type="album,artist,playlist")

        try:
            list_info = result["albums"]["items"][0]
            self._spotify_type = "album"
        except IndexError:
            try:
                list_info = result["artists"]["items"][0]
                self._spotify_type = "artist"
            except IndexError:
                try:
                    list_info = result["playlists"]["items"][0]
                    self._spotify_type = "playlist"
                except IndexError:
                    self.error = "No playlist found"
                    return None

        list_id = list_info["id"]
        self.title = list_info["name"]

        return list_id

    def fetch_metadata(self) -> bool:
        if self.title is None:
            if self._spotify_type == "playlist":
                result = self.api.playlist(self.id, fields="name")
            elif self._spotify_type == "artist":
                result = self.api.artist(self.id, fields="name")
            elif self._spotify_type == "album":
                result = self.api.album(self.id, fields="name")
            else:
                assert False
            self.title = result["name"]

        # download at most 50 tracks for a playlist (spotifys maximum)
        # for more tracks paging would need to be implemented
        if self._spotify_type == "playlist":
            result = self.api.playlist_tracks(
                self.id,
                fields="items(track(external_urls(spotify)))",
                limit=storage.get("max_playlist_items"),
            )
            track_infos = result["items"]
            for track_info in track_infos:
                try:
                    self.urls.append(track_info["track"]["external_urls"]["spotify"])
                except KeyError:
                    # skip songs that have no urls
                    pass
        elif self._spotify_type == "artist":
            result = self.api.artist_top_tracks(
                self.id, limit=storage.get("max_playlist_items")
            )
            tracks = result["tracks"]
            for track in tracks:
                self.urls.append(track["external_urls"]["spotify"])
        elif self._spotify_type == "album":
            result = self.api.album_tracks(
                self.id,
                fields="items(external_urls(spotify))",
                limit=storage.get("max_playlist_items"),
            )
            tracks = result["items"]
            for track in tracks:
                self.urls.append(track["external_urls"]["spotify"])

        return True



"""This module handles the led strip."""
from typing import Tuple

from core import redis
from core.lights.device import Device


class Strip(Device):
    """This class provides an interface to control the led strip."""

    def __init__(self, manager) -> None:
        super().__init__(manager, "strip")
        self.monochrome = True

        try:
            # https://github.com/adafruit/Adafruit_CircuitPython_PCA9685/blob/main/examples/pca9685_simpletest.py
            from board import SCL, SDA
            import busio
            from adafruit_pca9685 import PCA9685
        except (ModuleNotFoundError, ImportError):
            # There is a problem with board.py installs, catch it with the ImportError
            return

        try:
            i2c_bus = busio.I2C(SCL, SDA)
            self.controller = PCA9685(i2c_bus)
            self.controller.frequency = 60

            self.initialized = True
            redis.put("strip_initialized", True)
        except ValueError:
            # LED strip is not connected
            return

    def set_color(self, color: Tuple[float, float, float]) -> None:
        """Sets the color of the strip to the given rgb triple."""
        if not self.initialized:
            return

        for channel, val in enumerate(color):
            dimmed_val = val * self.brightness
            scaled_val = round(dimmed_val * 4095)
            self.controller.channels[channel].duty_cycle = scaled_val

    def clear(self) -> None:
        """Turns off the strip by setting its color to black."""
        if not self.initialized:
            return

        for channel in range(3):
            self.controller.channels[channel].duty_cycle = 0



from typing import Literal, overload

from cachetools import TTLCache

# Sometimes the storage functions are accessed dynamically.
# Comfort mypy by telling it the value will still be one of the specified ones.

class Interactivity:
        full_control = "Full Public Control"
        full_voting = "Up- and Downvoting"
        upvotes_only = "Upvotes Only"
        no_control = "No Control"

class Privileges:
        everybody = "Everybody"
        mod = "Mod and Admin"
        admin = "Admin Only"
        nobody = "Nobody"

PlatformEnabled = Literal[
    "local_enabled",
    "youtube_enabled",
    "spotify_enabled",
    "soundcloud_enabled",
    "jamendo_enabled",
]

PlatformSuggestions = Literal[
    "youtube_suggestions",
    "spotify_suggestions",
    "soundcloud_suggestions",
    "jamendo_suggestions",
]

DeviceBrightness = Literal[
    "ring_brightness", "strip_brightness", "wled_brightness", "screen_brightness"
]
DeviceMonochrome = Literal[
    "ring_monochrome", "strip_monochrome", "wled_monochrome", "screen_monochrome"
]
DeviceProgram = Literal[
    "ring_program",
    "last_ring_program",
    "strip_program",
    "last_strip_program",
    "wled_program",
    "last_wled_program",
    "screen_program",
    "last_screen_program",
]

cache: TTLCache = TTLCache(ttl=10, maxsize=128)

@overload
def get(
    key: Literal[
        "ip_checking",
        "logging_enabled",
        "hashtags_active",
        "privileged_stream",
        "online_suggestions",
        "new_music_only",
        "enqueue_first",
        "local_enabled",
        "youtube_enabled",
        "spotify_enabled",
        "soundcloud_enabled",
        "jamendo_enabled",
        "feed_cava",
        "paused",
        "shuffle",
        "repeat",
        "autoplay",
        "ring_monochrome",
        "strip_monochrome",
        "wled_monochrome",
        "screen_monochrome",
        "dynamic_resolution",
    ]
) -> bool: ...
@overload
def get(
    key: Literal[
        "next_color_index",
        "downvotes_to_kick",
        "number_of_suggestions",
        "max_playlist_items",
        "max_queue_length",
        "people_to_party",
        "youtube_suggestions",
        "spotify_suggestions",
        "jamendo_suggestions",
        "soundcloud_suggestions",
        "wled_led_count",
        "wled_port",
    ]
) -> int: ...
@overload
def get(
    key: Literal[
        "color_offset",
        "song_cooldown",
        "max_download_size",
        "alarm_probability",
        "buzzer_cooldown",
        "buzzer_success_probability",
        "volume",
        "ups",
        "program_speed",
        "ring_brightness",
        "strip_brightness",
        "wled_brightness",
        "screen_brightness",
    ]
) -> float: ...
@overload
def get(
    key: Literal[
        "interactivity",
        "color_indication",
        "connectivity_host",
        "additional_keywords",
        "forbidden_keywords",
        "spotify_username",
        "spotify_password",
        "spotify_device_client_id",
        "spotify_device_client_secret",
        "spotify_mopidy_client_id",
        "spotify_mopidy_client_secret",
        "spotify_redirect_uri",
        "spotify_authorized_url",
        "spotipy_token_info",
        "soundcloud_auth_token",
        "jamendo_client_id",
        "output",
        "backup_stream",
        "wled_ip",
        "ring_program",
        "last_ring_program",
        "strip_program",
        "last_strip_program",
        "wled_program",
        "last_wled_program",
        "screen_program",
        "last_screen_program",
    ]
) -> str: ...
@overload
def get(key: Literal["initial_resolution"]) -> tuple[int, int]: ...
@overload
def get(key: Literal["fixed_color"]) -> tuple[float, float, float]: ...
@overload
def put(
    key: Literal[
        "ip_checking",
        "logging_enabled",
        "hashtags_active",
        "privileged_stream",
        "online_suggestions",
        "new_music_only",
        "enqueue_first",
        "local_enabled",
        "youtube_enabled",
        "spotify_enabled",
        "soundcloud_enabled",
        "jamendo_enabled",
        "feed_cava",
        "paused",
        "shuffle",
        "repeat",
        "autoplay",
        "ring_monochrome",
        "strip_monochrome",
        "wled_monochrome",
        "screen_monochrome",
        "dynamic_resolution",
    ],
    value: bool,
) -> None: ...
@overload
def put(
    key: Literal[
        "next_color_index",
        "downvotes_to_kick",
        "number_of_suggestions",
        "max_playlist_items",
        "max_queue_length",
        "people_to_party",
        "youtube_suggestions",
        "spotify_suggestions",
        "jamendo_suggestions",
        "soundcloud_suggestions",
        "wled_led_count",
        "wled_port",
    ],
    value: int,
) -> None: ...
@overload
def put(
    key: Literal[
        "color_offset",
        "song_cooldown",
        "max_download_size",
        "alarm_probability",
        "buzzer_cooldown",
        "buzzer_success_probability",
        "volume",
        "ups",
        "program_speed",
        "ring_brightness",
        "strip_brightness",
        "wled_brightness",
        "screen_brightness",
    ],
    value: float,
) -> None: ...
@overload
def put(
    key: Literal[
        "interactivity",
        "color_indication",
        "connectivity_host",
        "additional_keywords",
        "forbidden_keywords",
        "spotify_username",
        "spotify_password",
        "spotify_device_client_id",
        "spotify_device_client_secret",
        "spotify_mopidy_client_id",
        "spotify_mopidy_client_secret",
        "spotify_redirect_uri",
        "spotify_authorized_url",
        "spotipy_token_info",
        "soundcloud_auth_token",
        "jamendo_client_id",
        "output",
        "backup_stream",
        "wled_ip",
        "ring_program",
        "last_ring_program",
        "strip_program",
        "last_strip_program",
        "wled_program",
        "last_wled_program",
        "screen_program",
        "last_screen_program",
    ],
    value: str,
) -> None: ...
@overload
def put(key: Literal["initial_resolution"], value: tuple[int, int]) -> None: ...
@overload
def put(key: Literal["fixed_color"], value: tuple[float, float, float]) -> None: ...
